'''gtf2gtf.py - manipulate transcript models
============================================

:Tags: Genomics Genesets GTF Manipulation

Purpose
-------

This script reads a gene set in :term:`gtf` format from stdin, applies some
transformation, and outputs a new gene set in :term:`gtf` format to stdout.
The transformation is chosen by the ``--method`` command line option.

Transformations available for use in this script can broadly be
classified into four categories:

1. sorting gene sets
2. manipulating gene models
3. filtering gene sets
4. setting/resetting fields within a gtf file

Further options for working with gtf files are available in gff2gff.py,
which can be run with the specification --is-gtf


Sorting gene sets
+++++++++++++++++

``sort``

   Sorts entries in gtf file by one or more fields

   +-----------------+---------------------------------------+
   | option          | order in which fields are sorted      |
   +-----------------|---------------------------------------+
   | gene            | gene_id, contig, start                |
   +-----------------+---------------------------------------+
   | gene+transcript | gene_id, transcript_id, contig, start |
   +-----------------+---------------------------------------+
   | contig+gene     | contig, gene_id, transcript_id, start |
   +-----------------+---------------------------------------+
   | transcript      | transcript_id, contig, start          |
   +-----------------+---------------------------------------+
   | position        | contig, start                         |
   +-----------------+---------------------------------------+
   | position+gene   | contig( gene_id, start )              |
   +-----------------+---------------------------------------+
   | gene+position   | gene_id, contig, start                |
   +-----------------+---------------------------------------+
   | gene+exon       | gene_id, exon_id                      |
   +-----------------+---------------------------------------+

   N.B. position+gene sorts by gene_id, start, then subsequently sorts
   flattened gene lists by contig, start


Manipulating gene-models
++++++++++++++++++++++++

Options that can be used to alter the features represented in a :term:`gtf`
file. Only one method can be specified at once.

Input gtfs need to be sorted so that features for a gene or transcript
appear consecutively within the file. This can be achevied using
``--method=sort``.


``genes-to-unique-chunks```
    Divide the complete length of a gene up into chunks that represent
    ranges of bases that are all present in the same set of transcripts.
    E.g. for two overlapping exons an entry will be output representing
    the overlap and a seperate entry each for the sequences only present
    in one. Ranges which are between the first TSS and last TTS but not
    present in any transcript (i.e. merged introns) are also output.
    Useful for DEXSeq like splicing analysis

``find-retained-introns``
    Finds intervals within a transcript that represent retained-introns,
    here a retained intron is considered to be an intron in one transcript
    that is entirely contianed within the exon of another. The retained
    intron will be assigned to the transcript with the containing exon. Where
    multiple, overlapping introns are contained within a single exon of a
    transcript, the union of the introns will be output. Thus when considering
    an indevidual transcript, outputs will be non-overlapping. However,
    overlapping, or even identical feature can be output if they belong to
    different transcripts.

``merge-exons``
    Merges overlapping exons for all transcripts of a gene, outputting
    the merged exons. Can be used in conjunction with
    ``merge-exons-distance`` to set the minimum distance that may
    appear between two exons before they are merged.If
    ``--mark-utr`` is set, the UTR regions will be output separately.

``merge-transcripts``
    Merges all transcripts of a gene. Outputs contains a single interval that
    spans the original gene (both introns and exons). If ``--with-utr`` is
    set, the output interval will also contain UTR.

``merge-genes``

    Merges genes that have overlapping exons, outputting a single
    gene_id and transcript_id for all exons of overlapping genes. The
    input needs te sorted by transcript " (Does not merge intervals on
    different strands).

``join-exons``
    Joins together all exons of a transcript, outputting a single
    interval that spans the original transcript (both introns and
    exons). Input needs to be sorted by transcript.

``intersect-transcripts``
    Finds regions representing the intersection of all transcripts of a gene.
    Output will contain intervals spanning only those bases covered by all
    transcripts. If ``--with-utr`` is set, the UTR will also be included in the
    intersect. This method only uses ``exon`` or ``CDS`` features.

``merge-introns``
    Outputs a single interval that spans the region between the start
    of the first intron and the end of last intron. Single exons genes
    will not be output. The input needs to be sorted by gene

``exons2introns``
    Merges overlapping introns for all transcripts of a gene,
    outputting the merged introns. Use ``--intron-min-length`` to
    ignore merged introns below a specified length. Use
    ``--intron-border`` to specify a number of residues to remove at
    either end of output introns (residues are removed prior to
    filtering on size when used in conjunction with
    ``--intron-min-length``).

``transcripts2genes``
    Cluster transcripts into genes by exon overlap ignoring any
    gene_ids in the :term:`gtf` file. May be used in conjunction with
    ``reset-strand``

The option ``permit-duplicates`` may be specified in order to
allow gene-ids to be duplicated within the input :term:`gtf` file
(i.e. for the same gene-id to appear non-consecutively within the
input file). However, this option currently only works for
``merge-exons``, ``merge-transcripts``, ``merge-introns``, and
``intersect-transcripts``. It DOES NOT work for ``merge-genes``,
``join-exons``, or ``exons-file2introns``.

Filtering gene sets
+++++++++++++++++++

Options that can be used to filter :term:`gtf` files. For further
detail see command line options.

Input gtfs need to be sorted so that features for a gene or transcript
appear consecutively within the file. This can be achevied using
``--method=sort --sort-order``.

``filter``
    When filtering on the basis of 'gene-id' or 'transcript-id' a
    filename containing ids to be removed may provided using
    ``--map-tsv-file``. Alternatively, a random subsample of
    genes/transcripts may be retained using
    ``--sam-fileple-size``. Use ``--min-exons-length`` in conjunction
    with ``--sam-fileple-size`` to specify a minimum length for
    genes/transcripts to be retained. Use ``--ignore-strand`` to set
    strand to '.' in output.

    Other filter options include longest-gene, longest-transcript,
    or representative-transcript.

    When filtering on the basis of gene-id, transcript-id or longest-gene,
    ``--invert-filter`` may be used to invert the selection.

``remove-overlapping``
    Given a second :term:`gff` formatted file (``--file-gff``) removes
    any features overlapping. Any transcripts that intersect intervals
    in the supplied file are removed.  (Does not account for strand.)

``remove-duplicates``
    Remove duplicate features from :term:`gtf` file. The type of
    feature to be removed is set by the option ``-duplicate-feature``.
    Setting ``--duplicate-feature`` to 'gene', 'transcript', or
    'coordinates' will remove any interval for which non-consecutive
    occurrances of specified term appear in input :term:`gtf` file.
    Setting to 'ucsc', will remove any interval for which
    transcript-id contains '_dup'.


Setting fields
++++++++++++++

Options for altering fields within :term:`gtf`.

``rename-genes``
    With a mapping file is provided using ``--map-tsv-file``, renames
    the gene_id to the one supplied. Outputs a :term:`gtf` file with
    field renamed. Any entry in input :term:`gtf` not appearing in
    mapping file is discarded.

``rename-transcripts``
    as ``rename-genes``, but renames the transcript_id.

``add-protein-id``
    Takes a map of transcript_id to protein_id from the a tsv file
    (see option ``--map-tsv-file``) and appends the protein_id
    provided to the attributes field.  Any entry with a transcript_id
    not appearing in the tsv file is discarded.

``renumber-genes``
    Renumber genes from 1 using the pattern provided in
    ``--pattern-identifier``.

``renumber-transcripts``
    Renumber transcripts from 1 using the pattern provided in
    ``--pattern-identifier``.

``unset-genes``
    Renumber genes from 1 using the pattern provided in
    ``--pattern-identifier``. Transcripts with the same gene-id in the
    input :term:`gtf` file will have different gene-ids in the output
    :term:`gtf` file.

``set-transcript-to-gene``
    Will set the transcript-id to the gene-id for each feature.

``set-gene-to-transcript``
    Will set the gene-id to the transcript-id for each each feature.

``set-protein-to-transcript``
    Will append transcript_id to attributes field as 'protein_id'

``set-score-to-distance``
    Will reset the score field (field 6) of each feature in input
    :term:`gtf` to be the distance from transcription start site to
    the start of the feature.  (Assumes input file is sorted by
    transcript-id)

``set-gene_biotype-to-source``
    Sets the ``gene_biotype`` attribute from the source column. Will only set
    if biotype attribute is not present in the current record.

``rename-duplicates``
    Rename duplicate gene_ids and transcript_ids by addition of
    numerical suffix

``set-source-to-transcript_biotype``
    Sets the source attribute to the ``transcript_biotype``
    attribute. Will only set if ``transcript_biotype`` attribute is
    present in the current record.

Usage
-----

The following example sorts the input gene set by gene
(``method=sort``) so that it can be used as input for
``method=intersect-transcripts`` that outputs genomic the genomic
regions within a gene that is covered by all transcripts in a gene.
Finally, the resultant transcripts are renamed with the pattern
"MERGED_%i"::

    cgat gtf2gtf
            --method=sort
            --sort-order=gene \
    | cgat gtf2gtf
               --method=intersect-transcripts
               --with-utr
    | cgat gtf2gtf
               --method=renumber-transcripts
               --pattern-identifier=MERGED_%i

Type::

    cgat gtf2gtf --help

for command line options.

Command line Options
--------------------

'''
import sys
import re
import random
import collections
import itertools

import CGAT.GTF as GTF
import CGAT.Experiment as E
import CGAT.Genomics as Genomics
import CGAT.Intervals as Intervals
import CGAT.IOTools as IOTools


def find_retained_introns(gene):
    '''Given a bundle of transcripts, find intervals matching retained
    introns. A retained intron is defined as an interval from an exon/intron
    boundary to the next where both boundaries are in the same exon of another
    transcript'''

    intron_intervals = [GTF.toIntronIntervals(transcript)
                        for transcript in gene]
    intron_intervals = list(set(
        itertools.chain.from_iterable(intron_intervals)))
    intron_intervals.sort()

    for transcript in gene:
        exons = iter(sorted(GTF.asRanges(transcript)))
        introns = iter(intron_intervals)
        retained_introns = []
        try:
            intron = next(introns)
            exon = next(exons)
            while True:

                if exon[1] < intron[0]:

                    exon = next(exons)
                    continue

                if intron[0] >= exon[0] and intron[1] <= exon[1]:
                    E.debug("exon %s of transcript %s contains intron %s" %
                            (exon, transcript[0].transcript_id, intron))
                    retained_introns.append(intron)
                intron = next(introns)
        except StopIteration:
            pass

        retained_introns = Intervals.combine(retained_introns)

        for intron in retained_introns:
            entry = GTF.Entry()
            entry = entry.copy(transcript[0])
            entry.start = intron[0]
            entry.end = intron[1]
            yield entry


def gene_to_blocks(gene):
    '''Given a bundle of all exons in a gene, create a seperate exon
    for each unqiue part of a exon, as well as one for introns. '''

    exons = [(e.start, e.end)
             for e in gene if e.feature == "exon"]

    exons = list(set(sum(exons, ())))
    exons.sort()

    entry = GTF.Entry()
    entry = entry.copy(gene[0])
    entry.transcript_id = "merged"
    entry.feature = "exon"
    entry.source = "merged"

    for i in range(len(exons)-1):
        entry.start = exons[i]
        entry.end = exons[i+1]
        entry.attributes["exon_id"] = str(i + 1)
        yield entry


def main(argv=None):

    if not argv:
        argv = sys.argv

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("--merge-exons-distance",
                      dest="merge_exons_distance",
                      type="int",
                      help="distance in nucleotides between "
                      "exons to be merged [%default].")

    parser.add_option("--pattern-identifier", dest="pattern", type="string",
                      help="pattern to use for renaming genes/transcripts. "
                      "The pattern should contain a %i, for example "
                      "--pattern-identifier=ENSG%010i [%default].")

    parser.add_option("--sort-order",
                      dest="sort_order",
                      type="choice",
                      choices=("gene",
                               "gene+transcript",
                               "transcript",
                               "position",
                               "contig+gene",
                               "position+gene",
                               "gene+position",
                               "gene+exon"),
                      help="sort input data [%default].")

    parser.add_option("--mark-utr",
                      dest="mark_utr",
                      action="store_true",
                      help="mark utr for method --merge-exons. "
                      "[%default].")

    parser.add_option(
        "--without-utr",
        dest="with_utr",
        action="store_false",
        help="exclude UTR in methods --merge-exons, merge-transcripts "
        "and intersect-transripts. Setting this option will remove "
        "non-coding transcripts. "
        "[%default].")

    parser.add_option(
        "--filter-method", dest="filter_method",
        type="choice",
        choices=("gene",
                 "transcript",
                 "longest-gene",
                 "longest-transcript",
                 "representative-transcript",
                 "proteincoding",
                 "lincrna"),
        help="Filter method to apply. Available filters are: "
        "'gene': filter by gene_id given in ``--map-tsv-file``, "
        "'transcript': filter by transcript_id given in ``--map-tsv-file``, "
        "'longest-gene': output the longest gene for overlapping genes ,"
        "'longest-transcript': output the longest transcript per gene,"
        "'representative-transcript': output the representative transcript "
        "per gene. The representative transcript is the transcript "
        "that shares most exons with other transcripts in a gene. "
        "The input needs to be sorted by gene. "
        "'proteincoding': only output protein coding features. "
        "'lincrna': only output lincRNA features. "
        "[%default].")

    parser.add_option("-a", "--map-tsv-file", dest="filename_filter",
                      type="string",
                      metavar="tsv",
                      help="filename of ids to map/filter [%default].")

    parser.add_option(
        "--gff-file", dest="filename_gff", type="string",
        metavar="GFF",
        help="second filename of features (see --remove-overlapping) "
        "[%default]")

    parser.add_option("--invert-filter",
                      dest="invert_filter",
                      action="store_true",
                      help="when using --filter, invert selection "
                      "(like grep -v). "
                      "[%default].")

    parser.add_option("--sample-size", dest="sample_size", type="int",
                      help="extract a random sample of size # if the option "
                      "'--method=filter --filter-method' is set "
                      "[%default].")

    parser.add_option(
        "--intron-min-length",
        dest="intron_min_length", type="int",
        help="minimum length for introns (for --exons-file2introns) "
        "[%default].")

    parser.add_option("--min-exons-length",
                      dest="min_exons_length",
                      type="int",
                      help="minimum length for gene (sum of exons) "
                      "(--sam-fileple-size) [%default].")

    parser.add_option(
        "--intron-border",
        dest="intron_border",
        type="int",
        help="number of residues to exclude at intron at either end "
        "(--exons-file2introns) [%default].")

    parser.add_option("--ignore-strand",
                      dest="ignore_strand",
                      action="store_true",
                      help="remove strandedness of features (set to '.') when "
                      "using ``transcripts2genes`` or ``filter``"
                      "[%default].")

    parser.add_option("--permit-duplicates", dest="strict",
                      action="store_false",
                      help="permit duplicate genes. "
                      "[%default]")

    parser.add_option(
        "--duplicate-feature",
        dest="duplicate_feature",
        type="choice",
        choices=("gene", "transcript", "both", "ucsc", "coordinates"),
        help="remove duplicates by gene/transcript. "
        "If ``ucsc`` is chosen, transcripts ending on _dup# are "
        "removed. This is necessary to remove duplicate entries "
        "that are next to each other in the sort order "
        "[%default]")

    parser.add_option("--use-gene-id", dest="use_geneid", action="store_true",
                      help="when merging transcripts, exons or introns, use "
                      "the parent gene_id as the transcript id.")

    parser.add_option("-m", "--method", dest="method", type="choice",
                      action="append",
                      choices=(
                          "add-protein-id",
                          "exons2introns",
                          "filter",
                          "find-retained-introns",
                          "genes-to-unique-chunks",
                          "intersect-transcripts",
                          "join-exons",
                          "merge-exons",
                          "merge-transcripts",
                          "merge-genes",
                          "merge-introns",
                          "remove-overlapping",
                          "remove-duplicates",
                          "rename-genes",
                          "rename-transcripts",
                          "rename-duplicates",
                          "renumber-genes",
                          "renumber-transcripts",
                          "set-transcript-to-gene",
                          "set-gene-to-transcript",
                          "set-protein-to-transcript",
                          "set-score-to-distance",
                          "set-gene_biotype-to-source",
                          "set-source-to-transcript_biotype",
                          "sort",
                          "transcript2genes",
                          "unset-genes"),
                      help="Method to apply [%default]."
                      "Please only select one.")

    parser.set_defaults(
        sort_order="gene",
        filter_method="gene",
        pattern="%i",
        merge_exons_distance=0,
        filename_filter=None,
        intron_border=None,
        intron_min_length=None,
        sample_size=0,
        min_exons_length=0,
        ignore_strand=False,
        mark_utr=False,
        with_utr=True,
        invert_filter=False,
        duplicate_feature=None,
        strict=True,
        method=None,
        use_geneid=False,
    )

    (options, args) = E.Start(parser, argv=argv)

    ninput, noutput, nfeatures, ndiscarded = 0, 0, 0, 0

    if options.method is None:
        raise ValueError("please specify a --method")

    if len(options.method) > 1:
        raise ValueError("multiple --method arguements specified")
    else:
        options.method = options.method[0]

    if options.method == "set-transcript-to-gene":

        for gff in GTF.iterator(options.stdin):

            ninput += 1

            gff.setAttribute("transcript_id", gff.gene_id)
            options.stdout.write("%s\n" % str(gff))

            noutput += 1
            nfeatures += 1

    elif options.method == "set-gene_biotype-to-source":

        for gff in GTF.iterator(options.stdin):

            ninput += 1

            if "gene_biotype" not in gff.attributes:
                gff.setAttribute("gene_biotype", gff.source)

            options.stdout.write("%s\n" % str(gff))

            noutput += 1
            nfeatures += 1

    elif options.method == "set-source-to-transcript_biotype":

        for gff in GTF.iterator(options.stdin):

            ninput += 1

            try:
                gff.source = gff.transcript_biotype
            except AttributeError:
                pass

            options.stdout.write("%s\n" % str(gff))

            noutput += 1
            nfeatures += 1

    elif options.method == "remove-duplicates":

        counts = collections.defaultdict(int)

        if options.duplicate_feature == "ucsc":
            store = []
            remove = set()
            f = lambda x: x[0].transcript_id

            gffs = GTF.transcript_iterator(
                GTF.iterator(options.stdin), strict=False)
            outf = lambda x: "\n".join([str(y) for y in x])

            for entry in gffs:
                ninput += 1
                store.append(entry)
                id = f(entry)
                if "_dup" in id:
                    remove.add(re.sub("_dup\d+", "", id))
                    remove.add(id)

            for entry in store:
                id = f(entry)
                if id not in remove:
                    options.stdout.write(outf(entry) + "\n")
                    noutput += 1
                else:
                    ndiscarded += 1
                    E.info("discarded duplicates for %s" % (id))
        else:

            if options.duplicate_feature == "gene":
                gffs = GTF.gene_iterator(
                    GTF.iterator(options.stdin), strict=False)
                f = lambda x: x[0][0].gene_id
                outf = lambda x: "\n".join(
                    ["\n".join([str(y) for y in xx]) for xx in x])
            elif options.duplicate_feature == "transcript":
                gffs = GTF.transcript_iterator(
                    GTF.iterator(options.stdin), strict=False)
                f = lambda x: x[0].transcript_id
                outf = lambda x: "\n".join([str(y) for y in x])
            elif options.duplicate_feature == "coordinates":
                gffs = GTF.chunk_iterator(GTF.iterator(options.stdin))
                f = lambda x: x[0].contig + "_" + \
                    str(x[0].start) + "-" + str(x[0].end)
                outf = lambda x: "\n".join([str(y) for y in x])

            store = []

            for entry in gffs:
                ninput += 1
                store.append(entry)
                id = f(entry)
                counts[id] += 1

            # Assumes GTF file sorted by contig then start
            last_id = ""
            if options.duplicate_feature == "coordinates":
                for entry in store:
                    id = f(entry)
                    if id == last_id:
                        ndiscarded += 1
                        E.info("discarded duplicates for %s: %i" %
                               (id, counts[id]))
                    else:
                        options.stdout.write(outf(entry) + "\n")
                        noutput += 1
                    last_id = id

            else:
                for entry in store:
                    id = f(entry)
                    if counts[id] == 1:
                        options.stdout.write(outf(entry) + "\n")
                        noutput += 1
                    else:
                        ndiscarded += 1
                        E.info("discarded duplicates for %s: %i" %
                               (id, counts[id]))

    elif "sort" == options.method:

        for gff in GTF.iterator_sorted(GTF.iterator(options.stdin),
                                       sort_order=options.sort_order):
            ninput += 1
            options.stdout.write("%s\n" % str(gff))
            noutput += 1
            nfeatures += 1

    elif "set-gene-to-transcript" == options.method:

        for gff in GTF.iterator(options.stdin):

            ninput += 1

            gff.setAttribute("gene_id", gff.transcript_id)
            options.stdout.write("%s\n" % str(gff))

            noutput += 1
            nfeatures += 1

    elif "set-protein-to-transcript" == options.method:

        for gff in GTF.iterator(options.stdin):
            ninput += 1
            gff.setAttribute("protein_id", gff.transcript_id)
            options.stdout.write("%s\n" % str(gff))
            noutput += 1
            nfeatures += 1

    elif "add-protein-id" == options.method:

        transcript2protein = IOTools.readMap(
            IOTools.openFile(options.filename_filter, "r"))

        missing = set()
        for gff in GTF.iterator(options.stdin):
            ninput += 1
            if gff.transcript_id not in transcript2protein:
                if gff.transcript_id not in missing:
                    E.debug(
                        ("removing transcript '%s' due to "
                         "missing protein id") % gff.transcript_id)
                    missing.add(gff.transcript_id)
                ndiscarded += 1
                continue

            gff.setAttribute(
                "protein_id", transcript2protein[gff.transcript_id])
            options.stdout.write("%s\n" % str(gff))
            noutput += 1
            nfeatures += 1

        E.info("transcripts removed due to missing protein ids: %i" %
               len(missing))

    elif "join-exons" == options.method:

        for exons in GTF.transcript_iterator(GTF.iterator(options.stdin)):
            ninput += 1
            strand = Genomics.convertStrand(exons[0].strand)
            contig = exons[0].contig
            transid = exons[0].transcript_id
            geneid = exons[0].gene_id
            biotype = exons[0].source
            all_start, all_end = min([x.start for x in exons]), max(
                [x.end for x in exons])
            y = GTF.Entry()
            y.contig = contig
            y.source = biotype
            y.feature = "transcript"
            y.start = all_start
            y.end = all_end
            y.strand = strand
            y.transcript_id = transid
            y.gene_id = geneid
            options.stdout.write("%s\n" % str(y))

    elif "merge-genes" == options.method:
        # merges overlapping genes
        #
        gffs = GTF.iterator_sorted_chunks(
            GTF.flat_gene_iterator(GTF.iterator(options.stdin)),
            sort_by="contig-strand-start")

        def iterate_chunks(gff_chunks):

            last = next(gff_chunks)
            to_join = [last]

            for gffs in gff_chunks:
                d = gffs[0].start - last[-1].end

                if gffs[0].contig == last[0].contig and \
                   gffs[0].strand == last[0].strand:
                    assert gffs[0].start >= last[0].start, \
                        ("input file should be sorted by contig, strand "
                         "and position: d=%i:\nlast=\n%s\nthis=\n%s\n") % \
                        (d,
                         "\n".join([str(x) for x in last]),
                         "\n".join([str(x) for x in gffs]))

                if gffs[0].contig != last[0].contig or \
                        gffs[0].strand != last[0].strand or \
                        d > 0:
                    yield to_join
                    to_join = []

                last = gffs
                to_join.append(gffs)

            yield to_join
            raise StopIteration

        for chunks in iterate_chunks(gffs):
            ninput += 1
            if len(chunks) > 1:
                gene_id = "merged_%s" % chunks[0][0].gene_id
                transcript_id = "merged_%s" % chunks[0][0].transcript_id
                info = ",".join([x[0].gene_id for x in chunks])
            else:
                gene_id = chunks[0][0].gene_id
                transcript_id = chunks[0][0].transcript_id
                info = None

            intervals = []
            for c in chunks:
                intervals += [(x.start, x.end) for x in c]

            intervals = Intervals.combine(intervals)
            # take single strand
            strand = chunks[0][0].strand

            for start, end in intervals:
                y = GTF.Entry()
                y.fromGTF(chunks[0][0], gene_id, transcript_id)
                y.start = start
                y.end = end
                y.strand = strand

                if info:
                    y.addAttribute("merged", info)
                options.stdout.write("%s\n" % str(y))
                nfeatures += 1

            noutput += 1

    elif options.method == "renumber-genes":

        map_old2new = {}
        for gtf in GTF.iterator(options.stdin):
            ninput += 1
            if gtf.gene_id not in map_old2new:
                map_old2new[gtf.gene_id] = options.pattern % (
                    len(map_old2new) + 1)
            gtf.setAttribute("gene_id", map_old2new[gtf.gene_id])
            options.stdout.write("%s\n" % str(gtf))
            noutput += 1

    elif options.method == "unset-genes":

        map_old2new = {}
        for gtf in GTF.iterator(options.stdin):
            ninput += 1
            key = gtf.transcript_id
            if key not in map_old2new:
                map_old2new[key] = options.pattern % (len(map_old2new) + 1)
            gtf.setAttribute("gene_id", map_old2new[key])
            options.stdout.write("%s\n" % str(gtf))
            noutput += 1

    elif options.method == "renumber-transcripts":

        map_old2new = {}
        for gtf in GTF.iterator(options.stdin):
            ninput += 1
            key = (gtf.gene_id, gtf.transcript_id)
            if key not in map_old2new:
                map_old2new[key] = options.pattern % (
                    len(map_old2new) + 1)
            gtf.setAttribute("transcript_id", map_old2new[key])
            options.stdout.write("%s\n" % str(gtf))
            noutput += 1

    elif options.method == "transcripts2genes":

        transcripts = set()
        genes = set()
        ignore_strand = options.ignore_strand
        for gtfs in GTF.iterator_transcripts2genes(
                GTF.iterator(options.stdin)):

            ninput += 1
            for gtf in gtfs:
                if ignore_strand:
                    gtf.strand = "."
                options.stdout.write("%s\n" % str(gtf))
                transcripts.add(gtf.transcript_id)
                genes.add(gtf.gene_id)
                nfeatures += 1
            noutput += 1

        E.info("transcripts2genes: transcripts=%i, genes=%i" %
               (len(transcripts), len(genes)))

    elif options.method in ("rename-genes", "rename-transcripts"):

        map_old2new = IOTools.readMap(
            IOTools.openFile(options.filename_filter, "r"))

        if options.method == "rename-transcripts":
            is_gene_id = False
        elif options.method == "rename-genes":
            is_gene_id = True

        for gff in GTF.iterator(options.stdin):
            ninput += 1

            if is_gene_id:
                if gff.gene_id in map_old2new:
                    gff.setAttribute("gene_id", map_old2new[gff.gene_id])
                else:
                    E.debug("removing missing gene_id %s" % gff.gene_id)
                    ndiscarded += 1
                    continue

            else:
                if gff.transcript_id in map_old2new:
                    gff.setAttribute(
                        "transcript_id", map_old2new[gff.transcript_id])
                else:
                    E.debug("removing missing transcript_id %s" %
                            gff.transcript_id)
                    ndiscarded += 1
                    continue

            noutput += 1
            options.stdout.write("%s\n" % str(gff))

    elif options.method == "filter":

        keep_genes = set()
        if options.filter_method == "longest-gene":
            iterator = GTF.flat_gene_iterator(GTF.iterator(options.stdin))
            coords = []
            gffs = []
            for gff in iterator:
                gff.sort(key=lambda x: x.start)
                coords.append((gff[0].contig,
                               min([x.start for x in gff]),
                               max([x.end for x in gff]),
                               gff[0].gene_id))
                gffs.append(gff)
            coords.sort()

            last_contig = None
            max_end = 0
            longest_gene_id = None
            longest_length = None

            for contig, start, end, gene_id in coords:
                ninput += 1
                if contig != last_contig or start >= max_end:
                    if longest_gene_id:
                        keep_genes.add(longest_gene_id)
                    longest_gene_id = gene_id
                    longest_length = end - start
                    max_end = end
                else:
                    if end - start > longest_length:
                        longest_length, longest_gene_id = end - start, gene_id
                last_contig = contig
                max_end = max(max_end, end)

            keep_genes.add(longest_gene_id)
            invert = options.invert_filter
            for gff in gffs:
                keep = gff[0].gene_id in keep_genes

                if (keep and not invert) or (not keep and invert):
                    noutput += 1
                    for g in gff:
                        nfeatures += 1
                        options.stdout.write("%s\n" % g)
                else:
                    ndiscarded += 1
        elif options.filter_method in ("longest-transcript",
                                       "representative-transcript"):

            iterator = GTF.gene_iterator(GTF.iterator(options.stdin))

            def selectLongestTranscript(gene):
                r = []
                for transcript in gene:
                    transcript.sort(key=lambda x: x.start)
                    length = transcript[-1].end - transcript[0].start
                    r.append((length, transcript))
                r.sort()
                return r[-1][1]

            def selectRepresentativeTranscript(gene):
                '''select a representative transcript.

                The representative transcript represent the largest number
                of exons over all transcripts.
                '''
                all_exons = []
                for transcript in gene:
                    all_exons.extend([(x.start, x.end)
                                      for x in transcript
                                      if x.feature == "exon"])
                exon_counts = {}
                for key, exons in itertools.groupby(all_exons):
                    exon_counts[key] = len(list(exons))
                transcript_counts = []
                for transcript in gene:
                    count = sum([exon_counts[(x.start, x.end)]
                                 for x in transcript if x.feature == "exon"])
                    # add transcript id to sort to provide a stable
                    # segmentation.
                    transcript_counts.append((count,
                                              transcript[0].transcript_id,
                                              transcript))
                transcript_counts.sort()
                return transcript_counts[-1][-1]

            if options.filter_method == "longest-transcript":
                _select = selectLongestTranscript
            elif options.filter_method == "representative-transcript":
                _select = selectRepresentativeTranscript

            for gene in iterator:
                ninput += 1
                # sort in order to make reproducible which
                # gene is chosen.
                transcript = _select(sorted(gene))
                noutput += 1
                for g in transcript:
                    nfeatures += 1
                    options.stdout.write("%s\n" % g)

        elif options.filter_method in ("gene", "transcript"):

            if options.filename_filter:

                ids = IOTools.readList(
                    IOTools.openFile(options.filename_filter, "r"))
                E.info("read %i ids" % len(ids))

                ids = set(ids)
                by_gene = options.filter_method == "gene"
                by_transcript = options.filter_method == "transcript"
                invert = options.invert_filter

                ignore_strand = options.ignore_strand
                for gff in GTF.iterator(options.stdin):

                    ninput += 1

                    keep = False
                    if by_gene:
                        keep = gff.gene_id in ids
                    if by_transcript:
                        keep = gff.transcript_id in ids
                    if (invert and keep) or (not invert and not keep):
                        continue

                    if ignore_strand:
                        gff.strand = "."

                    options.stdout.write("%s\n" % str(gff))
                    nfeatures += 1
                    noutput += 1

            elif options.sample_size:

                if options.filter_method == "gene":
                    iterator = GTF.flat_gene_iterator(
                        GTF.iterator(options.stdin))
                elif options.filter_method == "transcript":
                    iterator = GTF.transcript_iterator(
                        GTF.iterator(options.stdin))
                if options.min_exons_length:
                    iterator = GTF.iterator_min_feature_length(
                        iterator,
                        min_length=options.min_exons_length,
                        feature="exon")

                data = [x for x in iterator]
                ninput = len(data)
                if len(data) > options.sample_size:
                    data = random.sample(data, options.sample_size)

                for d in data:
                    noutput += 1
                    for dd in d:
                        nfeatures += 1
                        options.stdout.write(str(dd) + "\n")

            else:
                assert False, "please supply either a filename "
                "with ids to filter with (--map-tsv-file) or a sample-size."

        elif options.filter_method in ("proteincoding", "lincrna",
                                       "processed-pseudogene"):
            # extract entries by transcript/gene biotype.
            # This filter uses a test on the source field (ENSEMBL pre v78)
            # a regular expression on the attributes (ENSEMBL >= v78).
            tag = {"proteincoding": "protein_coding",
                   "processed-pseudogene": "processed_pseudogene",
                   "lincrna": "lincRNA"}[options.filter_method]
            rx = re.compile('"%s"' % tag)
            if not options.invert_filter:
                f = lambda x: x.source == tag or rx.search(x.attributes)
            else:
                f = lambda x: x.source != tag and not rx.search(x.attributes)

            for gff in GTF.iterator(options.stdin):
                ninput += 1
                if f(gff):
                    options.stdout.write(str(gff) + "\n")
                    noutput += 1
                else:
                    ndiscarded += 1

    elif options.method == "exons2introns":

        for gffs in GTF.flat_gene_iterator(GTF.iterator(options.stdin)):

            ninput += 1

            cds_ranges = GTF.asRanges(gffs, "CDS")
            exon_ranges = GTF.asRanges(gffs, "exon")
            input_ranges = Intervals.combine(cds_ranges + exon_ranges)

            if len(input_ranges) > 1:
                last = input_ranges[0][1]
                output_ranges = []
                for start, end in input_ranges[1:]:
                    output_ranges.append((last, start))
                    last = end

                if options.intron_border:
                    b = options.intron_border
                    output_ranges = [(x[0] + b, x[1] - b)
                                     for x in output_ranges]

                if options.intron_min_length:
                    l = options.intron_min_length
                    output_ranges = [
                        x for x in output_ranges if x[1] - x[0] > l]

                for start, end in output_ranges:

                    entry = GTF.Entry()
                    entry.copy(gffs[0])
                    entry.clearAttributes()
                    entry.transcript_id = "merged"
                    entry.feature = "intron"
                    entry.start = start
                    entry.end = end
                    options.stdout.write("%s\n" % str(entry))
                    nfeatures += 1
                noutput += 1
            else:
                ndiscarded += 1

    elif options.method == "set-score-to-distance":

        for gffs in GTF.transcript_iterator(GTF.iterator(options.stdin)):
            ninput += 1
            strand = Genomics.convertStrand(gffs[0].strand)
            all_start, all_end = min([x.start for x in gffs]), max(
                [x.end for x in gffs])

            if strand != ".":
                t = 0
                if strand == "-":
                    gffs.reverse()
                for gff in gffs:
                    gff.score = t
                    t += gff.end - gff.start

                if strand == "-":
                    gffs.reverse()
            for gff in gffs:
                options.stdout.write("%s\n" % str(gff))
                nfeatures += 1
            noutput += 1

    elif options.method == "remove-overlapping":

        index = GTF.readAndIndex(
            GTF.iterator(IOTools.openFile(options.filename_gff, "r")))

        for gffs in GTF.transcript_iterator(GTF.iterator(options.stdin)):
            ninput += 1
            found = False
            for e in gffs:
                if index.contains(e.contig, e.start, e.end):
                    found = True
                    break

            if found:
                ndiscarded += 1
            else:
                noutput += 1
                for e in gffs:
                    nfeatures += 1
                    options.stdout.write("%s\n" % str(e))

    elif options.method == "intersect-transcripts":

        for gffs in GTF.gene_iterator(GTF.iterator(options.stdin),
                                      strict=options.strict):

            ninput += 1
            r = []
            for g in gffs:
                if options.with_utr:
                    ranges = GTF.asRanges(g, "exon")
                else:
                    ranges = GTF.asRanges(g, "CDS")
                r.append(ranges)

            result = r[0]
            for x in r[1:]:
                result = Intervals.intersect(result, x)

            entry = GTF.Entry()
            entry.copy(gffs[0][0])
            entry.clearAttributes()
            entry.transcript_id = "merged"
            entry.feature = "exon"
            for start, end in result:
                entry.start = start
                entry.end = end
                options.stdout.write("%s\n" % str(entry))
                nfeatures += 1

            noutput += 1

    elif "rename-duplicates" == options.method:
        # note: this will only rename entries with "CDS" in feature column

        assert options.duplicate_feature in ["gene", "transcript", "both"],\
            ("for renaming duplicates, --duplicate-feature must be set to one "
             "of 'gene', transcript' or 'both'")

        gene_ids = list()
        transcript_ids = list()
        gtfs = list()

        for gtf in GTF.iterator(options.stdin):
            gtfs.append(gtf)
            if gtf.feature == "CDS":
                gene_ids.append(gtf.gene_id)
                transcript_ids.append(gtf.transcript_id)

        dup_gene = [item for item in set(gene_ids) if gene_ids.count(item) > 1]
        dup_transcript = [item for item in set(transcript_ids)
                          if transcript_ids.count(item) > 1]

        E.info("Number of duplicated gene_ids: %i" % len(dup_gene))
        E.info("Number of duplicated transcript_ids: %i" % len(dup_transcript))

        gene_dict = dict(list(zip(dup_gene, ([0] * len(dup_gene)))))
        transcript_dict = dict(list(zip(dup_transcript,
                                        ([0] * len(dup_transcript)))))

        for gtf in gtfs:
            if gtf.feature == "CDS":
                if options.duplicate_feature in ["both", "gene"]:
                    if gtf.gene_id in dup_gene:
                        gene_dict[gtf.gene_id] = gene_dict[gtf.gene_id] + 1
                        # TS. patch until pysam.ctabixproxies.pyx bugfixed
                        gtf.attributes = gtf.attributes.strip()
                        gtf.setAttribute('gene_id',
                                         gtf.gene_id + "." +
                                         str(gene_dict[gtf.gene_id]))

                if options.duplicate_feature in ["both", "transcript"]:
                    if gtf.transcript_id in dup_transcript:
                        transcript_dict[gtf.transcript_id] = \
                            transcript_dict[gtf.transcript_id] + 1
                        # TS. patch until pysam.ctabixproxies.pyx bugfixed
                        gtf.attributes = gtf.attributes.strip()
                        gtf.setAttribute(
                            'transcript_id',
                            gtf.transcript_id + "." +
                            str(transcript_dict[gtf.transcript_id]))

            options.stdout.write("%s\n" % gtf)

    elif options.method in ("merge-exons",
                            "merge-introns",
                            "merge-transcripts"):
        for gffs in GTF.flat_gene_iterator(
                GTF.iterator(options.stdin),
                strict=options.strict):
            ninput += 1

            cds_ranges = GTF.asRanges(gffs, "CDS")
            exon_ranges = GTF.asRanges(gffs, "exon")
            # sanity checks
            strands = set([x.strand for x in gffs])
            contigs = set([x.contig for x in gffs])
            if len(strands) > 1:
                E.warn('Skipping gene %s on multiple strands' % gffs[0].gene_id)
                break
                raise ValueError(
                    "can not merge gene '%s' on multiple strands: %s" % (
                        gffs[0].gene_id, str(strands)))

            if len(contigs) > 1:
                raise ValueError(
                    "can not merge gene '%s' on multiple contigs: %s" % (
                        gffs[0].gene_id, str(contigs)))

            strand = Genomics.convertStrand(gffs[0].strand)
            utr_ranges = []

            if cds_ranges and options.mark_utr:
                cds_start, cds_end = cds_ranges[0][0], cds_ranges[-1][1]
                midpoint = (cds_end - cds_start) / 2 + cds_start

                utr_ranges = []
                for start, end in Intervals.truncate(exon_ranges, cds_ranges):
                    if end - start > 3:
                        if strand == ".":
                            feature = "UTR"
                        elif strand == "+":
                            if start < midpoint:
                                feature = "UTR5"
                            else:
                                feature = "UTR3"
                        elif strand == "-":
                            if start < midpoint:
                                feature = "UTR3"
                            else:
                                feature = "UTR5"
                        utr_ranges.append((feature, start, end))

            try:
                biotypes = [x["gene_biotype"] for x in gffs]
                biotype = ":".join(set(biotypes))
            except (KeyError, AttributeError):
                biotype = None

            def output_ranges(ranges, gffs, biotype=None,
                              use_geneid=False):
                result = []
                for feature, start, end in ranges:
                    entry = GTF.Entry()
                    entry.copy(gffs[0])
                    entry.clearAttributes()
                    entry.feature = feature
                    if use_geneid:
                        entry.transcript_id = entry.gene_id
                    else:
                        entry.transcript_id = "merged"
                    if biotype:
                        entry.addAttribute("gene_biotype", biotype)
                    entry.start = start
                    entry.end = end
                    result.append(entry)
                return result

            result = []

            if options.method == "merge-exons":
                if options.with_utr:
                    if options.mark_utr:
                        result.extend(output_ranges(utr_ranges, gffs, biotype,
                                                    options.use_geneid))
                        r = [("CDS", x, y) for x, y in
                             Intervals.combineAtDistance(
                                 cds_ranges, options.merge_exons_distance)]
                    else:
                        r = [("exon", x, y) for x, y in
                             Intervals.combineAtDistance(
                                 exon_ranges, options.merge_exons_distance)]
                else:
                    r = [("CDS", x, y) for x, y in
                         Intervals.combineAtDistance(
                             cds_ranges, options.merge_exons_distance)]

            elif options.method == "merge-transcripts":

                if options.with_utr:
                    r = [("exon", exon_ranges[0][0],
                          exon_ranges[-1][1])]
                elif cds_ranges:
                    r = [("exon", cds_ranges[0][0],
                          cds_ranges[-1][1])]
                else:
                    ndiscarded += 1
                    continue

            elif options.method == "merge-introns":

                if len(exon_ranges) >= 2:
                    r = [("exon",
                          exon_ranges[0][1],
                          exon_ranges[-1][0])]
                else:
                    ndiscarded += 1
                    continue

            result.extend(output_ranges(r, gffs, biotype, options.use_geneid))

            result.sort(key=lambda x: x.start)

            for x in result:
                options.stdout.write("%s\n" % str(x))
                nfeatures += 1
            noutput += 1

    elif options.method == "find-retained-introns":

        for gene in GTF.gene_iterator(GTF.iterator(options.stdin)):
            ninput += 1
            found_any = False
            for intron in find_retained_introns(gene):
                found_any = True
                options.stdout.write("%s\n" % str(intron))
                nfeatures += 1
            if found_any:
                noutput += 1

    elif options.method == "genes-to-unique-chunks":

        for gene in GTF.flat_gene_iterator(GTF.iterator(options.stdin)):
            ninput += 1
            for exon in gene_to_blocks(gene):
                options.stdout.write("%s\n" % str(exon))
                nfeatures += 1
            noutput += 1

    else:
        raise ValueError("unknown method '%s'" % options.method)

    E.info("ninput=%i, noutput=%i, nfeatures=%i, ndiscarded=%i" %
           (ninput, noutput, nfeatures, ndiscarded))
    E.Stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''IndexedFasta.py - fast random access in fasta files
===================================================

This module provides fast random access to :term:`fasta` formatted
files that have been previously indexed. The indexing can be done
either through the samtools faidx tool (accessible through pysam_) or
using the in-house methods implemented in this module.

The main class is :class:`IndexedFasta`. This is a factory function
that provides transparent access to both samtools or CGAT indexed
fasta files.  The basic usage to retrieve the sequence spanning the
region chr12:10,000-10,100 is::

   from IndexedFasta import IndexedFasta
   fasta = IndexedFasta("hg19")
   fasta.getSequence("chr12", "+", 10000, 10100)

To index a file, use the :mod:`scripts/index_fasta` command line utility or the
:func:`createDatabase` function::

   > python index_fasta.py hg19 chr*.fa

This module has some useful utility functions:

:func:`splitFasta`
   split a :term:`fasta` formatted file into smaller pieces.

:func:`parseCoordinates`
   parse a coordinate string in various formats

but otherwise the module contains a multitude of additional functions that are
only of internal use.

Reference
----------

'''
import os
import sys
import array
import string
import re
import struct
import math
import tarfile
import random
import zlib
import gzip
import tempfile
import io
from CGAT import Experiment as E
import CGAT.IOTools as IOTools
import CGAT.Genomics as Genomics
from CGAT.AString import AString
import pysam
from future.moves import dbm
from six import StringIO

IS_PY3 = sys.version_info.major >= 3


class Uncompressor:

    def __init__(self, filename, unmangler):
        self.mFile = open(filename, "rb")
        self.mUnMangler = unmangler

    def read(self, block_size, indices, start, end):
        """read an uncompressed block from start:end.

        The compressed chunk starts at first_pos.
        NOTE: This is poorly implemented - do better.
        """

        # skip over uncompressed blocks
        d = int(math.floor(float(start) / block_size))
        r = start % block_size
        assert(d < len(indices))
        self.mFile.seek(indices[d])

        # read x bytes of compressed data, at least one full chunk.
        nchunks = int(math.ceil(float((r + end - start)) / block_size))

        fragments = []
        for x in range(d, d + nchunks):
            s = self.mFile.read(indices[x + 1] - indices[x])
            fragments.append(self.mUnMangler(s))
        u = "".join(fragments)

        assert len(u) >= end - start, \
            "fragment smaller than requested size: %i > %i-%i=%i" %\
            (len(u), end, start, end - start)

        return u[r:r + end - start]


def writeFragments(outfile_fasta,
                   outfile_index,
                   fragments,
                   mangler, size,
                   write_all=False):
    """write mangled fragments to *outfile_fasta* in chunks of *size*
    updating *outfile_index*.

    returns part of last fragment that has not been written and is
    less than *size* and the number of fragments output.

    If *write_all* is True, all of the fragments are written to
    the file and the last file position is added to *outfile_index*
    as well.
    """

    s = "".join(fragments)
    rest = len(s) % size
    if len(s) > size:
        for x in range(0, len(s) - rest, size):
            outfile_index.write("\t%i" % outfile_fasta.tell())
            outfile_fasta.write(mangler(s[x:x + size]))

    if rest:
        if write_all:
            outfile_index.write("\t%i" % outfile_fasta.tell())
            outfile_fasta.write(mangler(s[-rest:]))
            outfile_index.write("\t%i" % outfile_fasta.tell())
            return ""
        else:
            return s[-rest:]
    else:
        return ""


def gzip_mangler(s):

    xfile = StringIO()
    gzipfile = gzip.GzipFile(fileobj=xfile, mode="wb")
    gzipfile.write(s)
    gzipfile.close()

    m = xfile.getvalue()
    xfile.close()
    # write identifier

    return m


def gzip_demangler(s):
    if sys.version_info.major >= 3:
        gzipfile = io.TextIOWrapper(
            gzip.GzipFile(fileobj=io.BytesIO(s), mode="rb"),
            encoding="ascii")
    else:
        gzipfile = gzip.GzipFile(fileobj=io.BytesIO(s), mode="rb")
    m = gzipfile.readline()

    return m


class Translator:

    """translate a sequence."""

    def __init__(self):
        self.mRegEx = re.compile(" +")

    def __call__(self, sequence):
        return "".join(self.mMapScore2Char[self.mMapScore2Score[int(x)]]
                       for x in self.mRegEx.split(sequence.strip()))

    def translate(self, sequence):
        raise NotImplementedError("translate not implemented.")


class TranslatorPhred(Translator):

    """translate phred quality scores."""

    def __init__(self, *args, **kwargs):
        Translator.__init__(self, *args, **kwargs)
        self.mMapScore2Char = [chr(33 + x) for x in range(0, 93)]
        self.mMapScore2Score = list(range(0, 93))

    def translate(self, sequence):
        return array.array("I", (ord(x) - 33 for x in sequence))


class TranslatorSolexa(Translator):

    """translate solexa quality scores."""

    def __init__(self, *args, **kwargs):
        Translator.__init__(self, *args, **kwargs)
        self.mMapScore2Char = [chr(64 + x) for x in range(0, 128)]
        self.mMapScore2Score = [
            int(10.0 * math.log(1.0 + 10 ** (x / 10.0)) / math.log(10) + .499)
            for x in range(-64, 65)]

    def translate(self, sequence):
        raise NotImplementedError("translate not implemented.")
        # return array.array("i", (ord(x) - 64))


class TranslatorRange200(Translator):

    """translate pcap quality scores.

    For example for PCAP scores.

    These scores range from 0 to 100 and are the
    "a weighted sum of input base quality values
    (Huang and Madan 1999)

    The numerical values from 0 to 200 are stored
    as values form 33 to 233
    "
    """

    def __init__(self, *args, **kwargs):
        Translator.__init__(self, *args, **kwargs)
        self.mMapScore2Char = [chr(33 + x) for x in range(0, 200)]

    def __call__(self, sequence):
        try:
            return "".join(self.mMapScore2Char[int(x)]
                           for x in self.mRegEx.split(sequence.strip()))
        except ValueError as msg:
            raise ValueError(msg + " parsing error in fragment: %s" % sequence)

    def translate(self, sequence):
        return array.array("I", (ord(x) - 33 for x in sequence))


class TranslatorBytes(Translator):

    """output binary values as bytes permitting values from 0 to 255

    Note the resulting file will not be iterable as newline is not
    a record-separator any more.
    """

    def __init__(self, *args, **kwargs):
        Translator.__init__(self, *args, **kwargs)

    def __call__(self, sequence):
        try:
            return "".join(chr(int(x)) for x in
                           self.mRegEx.split(sequence.strip()))
        except ValueError as msg:
            print("parsing error in line: %s" % sequence)
            print("message=%s" % str(msg))
            return ""

    def translate(self, sequence):
        return array.array("I", (ord(x) for x in sequence))


class MultipleFastaIterator:

    def __init__(self,
                 filenames,
                 regex_identifier=None,
                 format="auto"):

        if isinstance(filenames, str):
            self.filenames = [filenames]
        else:
            self.filenames = filenames

        self.regexIdentifier = regex_identifier
        self.iterator = self._iterate()
        self.format = format

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return next(self.iterator)
        except StopIteration:
            return None

    def next(self):
        try:
            return next(self.iterator)
        except StopIteration:
            return None

    def _iterate(self):
        """iterate over muliple files."""

        def _iter(infile):

            identifier = None
            is_new = False

            for line in infile:
                if line.startswith("#"):
                    continue
                if line.startswith(">"):

                    if self.regexIdentifier:
                        try:
                            identifier = re.search(
                                self.regexIdentifier, line[1:-1]).groups()[0]
                        except AttributeError:
                            raise ValueError(
                                "could not parse identifier from line %s "
                                "- check the input" % line[1:-1])
                    else:
                        identifier = re.split("\s", line[1:-1])[0]
                    is_new = True
                else:
                    if not identifier:
                        raise ValueError(
                            "refusing to emit sequence without identifier "
                            "- check the input")
                    yield is_new, identifier, line.strip()
                    is_new = False

        for filename in self.filenames:
            if self.format == "tar.gz" or self.format == "tar" or \
               (self.format == "auto" and filename.endswith("tar.gz")):
                if filename == "-":
                    if IS_PY3:
                        tf = tarfile.open(fileobj=sys.stdin.buffer, mode="r|*")
                    else:
                        tf = tarfile.open(fileobj=sys.stdin, mode="r|*")
                else:
                    tf = tarfile.open(filename, mode="r")
                for f in tf:
                    b, ext = os.path.splitext(f.name)
                    if ext.lower() in (".fasta", ".fa"):
                        E.info("extracting %s" % f.name)
                        if sys.version_info.major >= 3:
                            infile = io.TextIOWrapper(
                                tf.extractfile(f),
                                encoding="ascii")
                        else:
                            infile = tf.extractfile(f)
                        for x in _iter(infile):
                            yield x
                    else:
                        E.info("skipping %s" % f.name)

                if tf != sys.stdin:
                    tf.close()
                continue
            elif self.format == "fasta.gz" or (self.format == "auto" and
                                               filename.endswith(".gz")):
                infile = IOTools.openFile(filename, "r")
            elif filename == "-":
                infile = sys.stdin
            else:
                infile = IOTools.openFile(filename, "r")

            for x in _iter(infile):
                yield x
            if filename != "-":
                infile.close()

        raise StopIteration


def createDatabase(db, iterator,
                   force=False,
                   synonyms=None,
                   compression=None,
                   random_access_points=None,
                   regex_identifier=None,
                   clean_sequence=False,
                   ignore_duplicates=False,
                   allow_duplicates=False,
                   translator=None):
    """index files in filenames to create database.

    Two new files are created - db.fasta and db_name.idx

    If compression is enabled, provide random access points
    every # bytes.

    Dictzip is treated as an uncompressed file.

    regex_identifier: pattern to extract identifier from description line.
    If None, the part until the first white-space character is used.

    translator: specify a translator
    """

    if db.endswith(".fasta"):
        db = db[:-len(".fasta")]

    if compression:
        if compression == "lzo":
            import lzo

            def lzo_mangler(s):
                return lzo.compress(s, 9)
            mangler = lzo_mangler
            db_name = db + ".lzo"
            write_chunks = True
        elif compression == "zlib":
            def zlib_mangler(s):
                return zlib.compress(s, 9)
            mangler = zlib_mangler
            db_name = db + ".zlib"
            write_chunks = True
        elif compression == "gzip":
            mangler = gzip_mangler
            db_name = db + ".gz"
            write_chunks = True
        elif compression == "dictzip":
            from . import dictzip

            def mangler(x):
                return x

            db_name = db + ".dz"
            write_chunks = False
        elif compression == "bzip2":
            import bz2

            def bzip_mangler(x):
                return bz2.compress(x, 9)

            mangler = bzip_mangler
            db_name = db + ".bz2"
            write_chunks = True
        elif compression == "debug":
            def mangler(x):
                return x
            db_name = db + ".debug"
            write_chunks = True
        elif compression == "rle":
            from . import RLE
            mangler = RLE.compress
            db_name = db + ".rle"
            write_chunks = True
        else:
            raise ValueError("unknown compression library: %s" % compression)

        index_name = db + ".cdx"

        if write_chunks and random_access_points is None \
           or random_access_points <= 0:
            raise ValueError("specify chunksize in --random-access-points")

    else:
        def mangler(x):
            return x
        db_name = db + ".fasta"
        write_chunks = False
        index_name = db + ".idx"

    if os.path.exists(db_name) and not force:
        raise ValueError("database %s already exists." % db_name)

    if os.path.exists(index_name) and not force:
        raise ValueError("database index %s already exists." % index_name)

    outfile_index = open(index_name, "w")
    if compression == "dictzip":
        if random_access_points is None or random_access_points <= 0:
            raise ValueError(
                "specify dictzip chunksize in --random-access-points")
        outfile_fasta = dictzip.open(
            db_name, "wb", buffersize=1000000, chunksize=random_access_points)
        compression = None
    else:
        outfile_fasta = open(db_name, "w")

    identifiers = {}
    lsequence = 0
    identifier_pos, sequence_pos = 0, 0

    if sys.version_info.major >= 3:
        translation = str.maketrans("xX", "nN")
    else:
        translation = string.maketrans("xX", "nN")

    fragments = []
    lfragment = 0

    last_identifier = None

    while 1:

        try:
            result = next(iterator)
        except StopIteration:
            break

        if not result:
            break

        is_new, identifier, fragment = result

        if is_new:
            # check for duplicate identifiers
            if identifier in identifiers:
                if ignore_duplicates:
                    raise ValueError("ignore duplicates not implemented")
                elif allow_duplicates:
                    # the current implementation will fail if the same
                    # identifiers
                    # are directly succeeding each other
                    # better: add return to iterator that indicates a new
                    # identifier
                    out_identifier = identifier + \
                        "_%i" % (identifiers[identifier])
                    identifiers[identifier] += 1
                    identifiers[out_identifier] = 1
                else:
                    raise ValueError("%s occurs more than once" %
                                     (identifier,))
            else:
                identifiers[identifier] = 1
                out_identifier = identifier

            if last_identifier:
                if write_chunks:
                    writeFragments(outfile_fasta, outfile_index,
                                   fragments, mangler,
                                   size=random_access_points,
                                   write_all=True)

                    fragments = []
                    lfragment = 0
                else:
                    outfile_fasta.write("\n")

                outfile_index.write("\t%i\n" % lsequence)

            identifier_pos = outfile_fasta.tell()
            outfile_fasta.write(mangler(">%s\n" % out_identifier))
            sequence_pos = outfile_fasta.tell()

            outfile_index.write("%s\t%i" % (out_identifier,
                                            identifier_pos))
            if write_chunks:
                outfile_index.write("\t%i" % random_access_points)
            else:
                outfile_index.write("\t%i" % sequence_pos)

            fragments = []
            lsequence = 0
            last_identifier = identifier

        if translator:
            s = translator(fragment)
        else:
            s = re.sub("\s", "", fragment.strip())
            if clean_sequence:
                s = s.translate(translation)

        lsequence += len(s)

        if write_chunks:
            fragments.append(s)
            lfragment += len(s)
            if lfragment > random_access_points:
                rest = writeFragments(outfile_fasta,
                                      outfile_index,
                                      fragments,
                                      mangler,
                                      size=random_access_points,
                                      write_all=False)
                fragments = [rest]
                lfragment = len(rest)
        else:
            outfile_fasta.write(mangler(s))

    if write_chunks:
        writeFragments(outfile_fasta, outfile_index, fragments, mangler,
                       size=random_access_points, write_all=True)
    else:
        outfile_fasta.write("\n")

    outfile_index.write("\t%i\n" % lsequence)

    # add synonyms for the table
    if synonyms:
        for key, vals in list(synonyms.items()):
            for val in vals:
                outfile_index.write("%s\t%s\n" % (key, val))

NAME_MAP = {
    'uncompressed': ('fasta', 'idx', False),
    'lzo': ('lzo',   'cdx', True),
    'dictzip': ('dz',    'idx', False),
    'zlib': ('zlib',  'cdx', True),
    'gzip': ('gz',  'cdx', True),
    'bzip2': ('bz2',   'cdx', True),
    'debug': ('debug', 'cdx', True),
}

PREFERENCES = (
    'uncompressed', 'lzo', 'dictzip', 'zlib', 'gzip', 'bzip2', 'debug')


class CGATIndexedFasta:

    """an indexed fasta file."""

    def __init__(self, dbname):

        if dbname.endswith(".fasta"):
            dbname = dbname[:-len(".fasta")]

        for x in PREFERENCES:
            d = "%s.%s" % (dbname, NAME_MAP[x][0])
            i = "%s.%s" % (dbname, NAME_MAP[x][1])
            if os.path.exists(d) and os.path.exists(i):
                self.mMethod = x
                self.mDbname = d
                self.mNameIndex = i
                self.mNoSeek = NAME_MAP[x][2]
                break
        else:
            raise KeyError("unknown database %s" % dbname)

        self.mIsLoaded = False
        self.mSynonyms = {}
        self.mConverter = None
        self.mIndex = {}
        self.mTranslator = None

    def __len__(self):
        """return the number of sequences in fasta file."""
        if not self.mIsLoaded:
            self._loadIndex()
        return len(self.mIndex)

    def __contains__(self, contig):
        if not self.mIsLoaded:
            self._loadIndex()
        return contig in self.mIndex or contig in self.mSynonyms

    def __getitem__(self, key):
        """return full length sequence."""
        return self.getSequence(key, "+", 0, 0, as_array=True)

    def _loadIndex(self, compress=False):
        """load complete index into memory.

        if compress is set to true, the index will not be loaded,
        but a compressed index will be created instead.
        """
        if self.mMethod == "uncompressed":
            self.mDatabaseFile = open(self.mDbname, "r")
        elif self.mMethod == "dictzip":
            from . import dictzip
            self.mDatabaseFile = dictzip.GzipFile(self.mDbname)
        elif self.mMethod == "lzo":
            import lzo
            self.mDatabaseFile = Uncompressor(self.mDbname, lzo.decompress)
        elif self.mMethod == "gzip":
            self.mDatabaseFile = Uncompressor(self.mDbname, gzip_demangler)
        elif self.mMethod == "zlib":
            self.mDatabaseFile = Uncompressor(self.mDbname, zlib.decompress)
        elif self.mMethod == "bzip2":
            import bz2
            self.mDatabaseFile = Uncompressor(self.mDbname, bz2.decompress)
        elif self.mMethod == "debug":
            self.mDatabaseFile = Uncompressor(
                self.mDbname + ".debug", lambda x: x)

        filename_index = self.mNameIndex + ".dbm"

        if compress:
            # if os.path.exists(filename_index):
            #     raise OSError("file %s already exists" % filename_index)
            self.mIndex = dbm.open(filename_index, "n")
        elif os.path.exists(filename_index):
            self.mIndex = dbm.open(filename_index, "r")
            if len(self.mIndex) == 0:
                raise ValueError(
                    "length of index is 0, possibly a python version problem")
            self.mIsLoaded = True
            return
        else:
            self.mIndex = {}

        for line in open(self.mNameIndex, "r"):

            data = line[:-1].split("\t")

            if len(data) == 2:
                # ignore synonyms of non-existent contigs
                identifier = data[1]
                if data[0] not in self.mIndex:
                    continue
                self.mSynonyms[identifier] = data[0]
            else:
                # index with random access points
                if len(data) > 4:
                    (identifier, pos_id, block_size, lsequence) = data[
                        0], int(data[1]), int(data[2]), int(data[-1])
                    points = list(map(int, data[3:-1]))
                    self.mIndex[identifier] = (
                        pos_id, block_size, lsequence, points)
                else:
                    (identifier, pos_id, pos_seq, lsequence) = data[
                        0], int(data[1]), int(data[2]), int(data[-1])
                    self.mIndex[identifier] = struct.pack(
                        "QQi", pos_id, pos_seq, lsequence)

        self._addSynonyms()
        self.mIsLoaded = True

    def _addSynonyms(self):
        '''add synonyms to indices.
        '''

        # Treat common cases of naming incompatibilites like chr1 = 1.
        # Truncate or add known prefixes
        def _add(src, target):
            for p in ("chr", "contig", "scaffold", "Chr"):
                if src.startswith(p):
                    k = src[len(p):]
                    if k not in self.mIndex:
                        self.mSynonyms[k] = target
                    # add lower/upper-case version
                    k = src[0].upper() + src[1:]
                    if k not in self.mIndex:
                        self.mSynonyms[k] = target
                    k = src[0].lower() + src[1:]
                    if k not in self.mIndex:
                        self.mSynonyms[k] = target
                    break
            else:
                for p in ("chr", "contig", "scaffold"):
                    k = "%s%s" % (p, src)
                    if k not in self.mIndex:
                        self.mSynonyms[k] = target

        k = list(self.mSynonyms.items())

        # fix the ambiguity between chrMT and chrM between UCSC and ENSEMBL
        if "chrM" in self.mIndex and "MT" not in self.mIndex:
            self.mSynonyms["MT"] = "chrM"
        elif "chrM" not in self.mSynonyms and "MT" in self.mSynonyms:
            self.mSynonyms["chrM"] = "MT"

        # this does the same thing for "chrMito" used in the yeast
        # genome
        if "chrM" in self.mIndex and "chrMito" not in self.mIndex:
            self.mSynonyms["chrMito"] = "chrM"

        for key, val in k:
            _add(key, val)

        # add pointers to self
        for key in list(self.mIndex.keys()):
            _add(key, key)

    def setTranslator(self, translator=None):
        """set the :class:`Translator` to use."""
        self.mTranslator = translator

    def getDatabaseName(self):
        """returns the name of the database."""
        return self.mDbname

    def getToken(self, contig):
        """check if token is in index."""
        if not self.mIsLoaded:
            self._loadIndex()

        # First Convert the contig to string to compare string with strings
        contig = str(contig)

        if contig in self.mSynonyms:
            contig = self.mSynonyms[contig]

        if contig not in self.mIndex:
            raise KeyError("%s not in index" % contig)

        return contig

    def getLength(self, contig):
        """return sequence length for sbjct_token."""
        if not self.mIsLoaded:
            self._loadIndex()
        data = self.mIndex[self.getToken(contig)]
        try:
            pos_id, pos_seq, lcontig = struct.unpack("QQi", data)
        except struct.error:
            pos_id, pos_seq, lcontig, points = data
        return lcontig

    def getLengths(self):
        """return all sequence lengths."""
        if not self.mIsLoaded:
            self._loadIndex()
        results = []
        for contig in list(self.mIndex.values()):
            data = self.mIndex[self.getToken(contig)]
            try:
                pos_id, pos_seq, lcontig = struct.unpack("QQi", data)
            except struct.error:
                pos_id, pos_seq, lcontig, points = data
            results.append(lcontig)
        return results

    def compressIndex(self):
        """compress index.
        Creates a database interface to an index.
        """
        self._loadIndex(compress=True)

    def getContigs(self):
        """return a list of contigs (no synonyms)."""
        if not self.mIsLoaded:
            self._loadIndex()
        return list(self.mIndex.keys())

    def getContigSizes(self, with_synonyms=True):
        """return hash with contig sizes including synonyms."""
        if not self.mIsLoaded:
            self._loadIndex()

        contig_sizes = {}
        for key, val in list(self.mIndex.items()):
            contig_sizes[key] = self.getLength(key)

        if with_synonyms:
            for key, val in list(self.mSynonyms.items()):
                contig_sizes[key] = self.getLength(val)

        return contig_sizes

    def setConverter(self, converter):
        """set converter from coordinate system to 0-based, both strand,
        open/closed coordinate system.

        """
        self.mConverter = converter

    def getSequence(self,
                    contig,
                    strand="+",
                    start=0,
                    end=0,
                    converter=None,
                    as_array=False):
        """get a genomic fragment.

        A genomic fragment is identified by the coordinates
        contig, strand, start, end.

        The converter function supplied translated these coordinates
        into 0-based coordinates. By default, start and end are assumed
        to be pythonic coordinates and are forward/reverse coordinates.

        If as_array is set to true, return the AString object. This might
        be beneficial for large sequence chunks. If as_array is set to False,
        return a python string.
        """

        contig = self.getToken(contig)

        data = self.mIndex[contig]
        # dummy is
        # -> pos_seq for seekable streams
        # -> block_size for unseekable streams
        try:
            pos_id, dummy, lsequence = struct.unpack("QQi", data)
        except (struct.error, TypeError):
            pos_id, dummy, lsequence, points = data

        pos_seq = dummy
        block_size = dummy

        if end == 0:
            end = lsequence

        if end > lsequence:
            raise ValueError(
                "3' coordinate on %s out of bounds: %i > %i" %
                (contig, end, lsequence))

        if start < 0:
            raise ValueError(
                "5' coordinate on %s out of bounds: %i < 0" % (contig, start))

        if converter:
            first_pos, last_pos = converter(start, end,
                                            str(strand) in ("+", "1"),
                                            lsequence)
        elif self.mConverter:
            first_pos, last_pos = self.mConverter(start, end,
                                                  str(strand) in ("+", "1"),
                                                  lsequence)
        else:
            first_pos, last_pos = start, end
            if str(strand) in ("-", "0", "-1"):
                first_pos, last_pos = lsequence - \
                    last_pos, lsequence - first_pos

        if first_pos == last_pos:
            return ""

        assert first_pos < last_pos, \
            "first position %i is larger than last position %i " % \
            (first_pos, last_pos)

        p = AString()

        if self.mNoSeek:
            # read directly from position
            p.fromstring(
                self.mDatabaseFile.read(block_size, data[3],
                                        first_pos, last_pos))
        else:
            first_pos += pos_seq
            last_pos += pos_seq

            self.mDatabaseFile.seek(first_pos)
            p.fromstring(self.mDatabaseFile.read(last_pos - first_pos))

        if str(strand) in ("-", "0", "-1"):
            p = AString(Genomics.complement(str(p)))

        if self.mTranslator:
            return self.mTranslator.translate(p)
        elif as_array:
            return p
        else:
            if IS_PY3:
                return p.tostring().decode("ascii")
            else:
                return p.tostring()

    def getRandomCoordinates(self, size):
        """returns coordinates for a random fragment of size #.

        The coordinates are forward/reverse.

        Default sampling mode:

        Each residue has the same probability of being
        in a fragment. Thus, the fragment can be smaller than
        size due to contig boundaries.
        """
        if not self.mIsLoaded:
            self._loadIndex()

        token = random.choice(list(self.mIndex.keys()))
        strand = random.choice(("+", "-"))
        data = self.mIndex[token]
        try:
            pos_id, pos_seq, lcontig = struct.unpack("QQi", data)
        except (struct.error, TypeError):
            pos_id, pos_seq, lcontig, points = data

        start, end = 0, 0
        if size >= lcontig:
            end = lcontig
        else:
            while end - start == 0:
                rpos = random.randint(0, lcontig)
                if random.choice(("True", "False")):
                    start = rpos
                    end = min(rpos + size, lcontig)
                else:
                    start = max(0, rpos - size)
                    end = rpos

        return token, strand, start, end


class PysamIndexedFasta(CGATIndexedFasta):

    '''interface a  pysam/samtools indexed fasta file with the
    CGATIndexedFasta API.'''

    def __init__(self, dbname):

        # open database file and truncate
        if os.path.exists(dbname) and dbname.endswith(".fa"):
            self.mDatabaseFile = pysam.Fastafile(dbname)
            dbname = dbname[:-len(".fa")]
        elif os.path.exists(dbname) and dbname.endswith(".fasta"):
            self.mDatabaseFile = pysam.Fastafile(dbname)
            dbname = dbname[:-len(".fasta")]
        elif os.path.exists(dbname + ".fa"):
            self.mDatabaseFile = pysam.Fastafile(dbname + ".fa")
        elif os.path.exists(dbname + ".fasta"):
            self.mDatabaseFile = pysam.Fastafile(dbname + ".fasta")

        if os.path.exists(dbname + ".fai"):
            self.mNameIndex = dbname + ".fai"
        elif os.path.exists(dbname + ".fa.fai"):
            self.mNameIndex = dbname + ".fa.fai"
        elif os.path.exists(dbname + ".fasta.fai"):
            self.mNameIndex = dbname + ".fasta.fai"
        else:
            raise ValueError("PysamIndexedFasta requires pre-built index")

        self.mMethod = "faidx"
        self.mDbname = dbname
        self.mNoSeek = False
        self.mIsLoaded = False
        self.mSynonyms = {}
        self.mConverter = None
        self.mIndex = {}
        self.mTranslator = None

    def _loadIndex(self, compress=False):
        '''load index into memory.'''

        with open(self.mNameIndex, "r") as infile:

            for line in infile:
                contig, lsequence, offset, line_blen, line_len = line[
                    :-1].split()
                self.mIndex[contig] = struct.pack(
                    "QQi", int(offset), int(offset), int(lsequence))

        self.mIsLoaded = True
        self._addSynonyms()

    def getSequence(self,
                    contig,
                    strand="+",
                    start=0,
                    end=0,
                    converter=None,
                    as_array=False):

        contig = self.getToken(contig)

        data = self.mIndex[contig]
        try:
            pos_id, pos_seq, lsequence = struct.unpack("QQi", data)
        except struct.error:
            pos_id, pos_seq, lsequence, points = data

        # convert to 0-based positive strand coordinates
        if converter:
            first_pos, last_pos = converter(start, end,
                                            str(strand) in ("+", "1"),
                                            lsequence)
        elif self.mConverter:
            first_pos, last_pos = self.mConverter(start, end,
                                                  str(strand) in ("+", "1"),
                                                  lsequence)
        else:
            first_pos, last_pos = start, end
            if str(strand) in ("-", "0", "-1"):
                first_pos, last_pos = lsequence - \
                    last_pos, lsequence - first_pos

        sequence = self.mDatabaseFile.fetch(contig, first_pos, last_pos)

        if str(strand) in ("-", "0", "-1"):
            try:
                # works in py2 only
                sequence = string.translate(
                    sequence[::-1],
                    string.maketrans("ACGTacgtNn", "TGCAtgcaNn"))
            except AttributeError:
                # works in py3 only
                sequence = str(sequence[::-1]).translate(
                    str.maketrans("ACGTacgtNn", "TGCAtgcaNn"))

        return sequence


def IndexedFasta(dbname, *args, **kwargs):
    '''factory function for IndexedFasta objects.'''

    if (os.path.exists(dbname) or os.path.exists(dbname + ".fa")) \
            and (os.path.exists(dbname + ".fai") or
                 os.path.exists(dbname + ".fa.fai")):
        return PysamIndexedFasta(dbname, *args, **kwargs)
    else:
        return CGATIndexedFasta(dbname, *args, **kwargs)


def _one_forward_closed(x, y, c, l):
    """convert coordinates to zero-based, both strand, open/closed coordinates.

     Parameters are from, to, is_positive_strand, length of contig.
    """
    x -= 1
    if not c:
        x, y = l - y, l - x
    return x, y


def _zero_forward_closed(x, y, c, l):
    """convert coordinates to zero-based, both strand, open/closed coordinates.

     Parameters are from, to, is_positive_strand, length of contig.
    """
    y += 1
    if not c:
        x, y = l - y, l - x
    return x, y


def _one_both_closed(x, y, c=None, l=None):
    """convert coordinates to zero-based, both strand, open/closed coordinates.

     Parameters are from, to, is_positive_strand, length of contig.
    """
    return x - 1, y


def _zero_both_closed(x, y, c=None, l=None):
    """convert coordinates to zero-based, both strand, open/closed coordinates.

     Parameters are from, to, is_positive_strand, length of contig.
    """
    return x, y + 1


def _one_forward_open(x, y, c, l):
    """convert coordinates to zero-based, both strand, open/closed coordinates.

     Parameters are from, to, is_positive_strand, length of contig.
    """
    x -= 1
    y -= 1
    if not c:
        x, y = l - y, l - x
    return x, y


def _zero_forward_open(x, y, c, l):
    """convert coordinates to zero-based, both strand, open/closed coordinates.

     Parameters are from, to, is_positive_strand, length of contig.
    """
    if not c:
        x, y = l - y, l - x
    return x, y


def _one_both_open(x, y, c=None, l=None):
    """convert coordinates to zero-based, both strand, open/closed coordinates.

     Parameters are from, to, is_positive_strand, length of contig.
    """
    return x - 1, y - 1


def _zero_both_open(x, y, c=None, l=None):
    """convert coordinates to zero-based, both strand, open/closed coordinates.

    Parameters are from, to, is_positive_strand, length of contig.
    """
    return x, y


def getConverter(format):
    """return a converter function for converting various
    coordinate schemes into 0-based, both strand, closed-open ranges.

    converter functions have the parameters
    x, y, s, l: with x and y the coordinates of
    a sequence fragment, s the strand (True is positive)
    and l being the length of the contig.

    Format is a "-" separated combination of the keywords
    "one", "zero", "forward", "both", "open", "closed":

    zero/one: zero or one-based coordinates
    forward/both: forward coordinates or forward/reverse coordinates
    open/closed: half-open intervals (pythonic) or closed intervals

    """

    data = set(format.split("-"))

    if "one" in data:
        if "forward" in data:
            if "closed" in data:
                return _one_forward_closed
            else:
                return _one_forward_open
        else:
            if "closed" in data:
                return _one_both_closed
            else:
                return _one_both_open
    else:
        if "forward" in data:
            if "closed" in data:
                return _zero_forward_closed
            else:
                return _zero_forward_open
        else:
            if "closed" in data:
                return _zero_both_closed
            else:
                return _zero_both_open


def benchmarkRandomFragment(fasta, size):
    """returns a random fragment of size."""

    contig, strand, start, end = fasta.getRandomCoordinates(size)
    s = fasta.getSequence(contig, strand, start, end)
    return s


def verify(fasta1, fasta2, num_iterations, fragment_size,
           stdout=sys.stdout, quiet=False):
    """verify two databases.

    Get segment from fasta1 and check for presence in fasta2.
    """
    if not quiet:
        stdout.write(
            "verifying %s and %s using %i random segments of length %i\n" %
            (fasta1.getDatabaseName(),
             fasta2.getDatabaseName(),
             num_iterations,
             fragment_size))
        stdout.flush()
    nerrors = 0
    for x in range(num_iterations):
        contig, strand, start, end = fasta1.getRandomCoordinates(fragment_size)
        s1 = fasta1.getSequence(contig, strand, start, end)
        s2 = fasta2.getSequence(contig, strand, start, end)
        if s1 != s2:
            if not quiet:
                stdout.write("discordant segment: %s:%s:%i:%i\n%s\n%s\n" %
                             (contig, strand, start, end, s1, s2))
            nerrors += 1
    return nerrors


def splitFasta(infile, chunk_size, dir="/tmp", pattern=None):
    """split a fasta file into a subset of files.

    If pattern is not given, random file names are chosen.
    """

    n = 0
    chunk = 0

    def _getFilename(chunk):
        if pattern:
            outname = pattern % chunk
            outfile = os.open(outname, "w")
        else:
            (outfile, outname) = tempfile.mkstemp(dir=dir)
        return (outfile, outname)

    outfile, outname = _getFilename(chunk)
    filenames = [outname]
    noutput = 0
    for line in infile:
        if line[0] == "#":
            continue
        if line[0] == ">":
            n += 1
            if n > chunk_size:
                os.close(outfile)
                n = 1
                outfile, outname = _getFilename(chunk)
                filenames.append(outname)
            noutput += 1

        os.write(outfile, line)

    os.close(outfile)

    if noutput == 0:
        os.remove(outname)
        return []

    return filenames


def parseCoordinates(s):
    '''parse a coordinate string.

    The coordinate string can be various formats, such as
    ``chr1:+:10:1000``, ``chr1:10..1000``.

    Returns
    -------
    contig : string
        The chromosome/contig.
    strand : char
        Strand. If not present, set to "+".
    start : int
        Start of interval
    end : int
        End of interval. If not present, set to start + 1.
    '''

    if ":" in s:
        d = s.strip().split(":")
        if len(d) == 4:
            contig, strand, start, end = d
        elif len(d) == 2:
            contig = d[0]
            strand = "+"
            if ".." in d[1]:
                start, end = d[1].split("..")
            else:
                start = d[1]
        else:
            raise ValueError("format not recognized")
    else:
        contig = s
        strand = "+"
        start = "0"
        # full sequence
        end = "0"

    start = int(re.sub(",", "", start))
    if end:
        end = int(re.sub(",", "", end))
    else:
        end = start + 1

    return contig, strand, start, end



'''
gff2coverage.py - compute genomic coverage of gff intervals
===========================================================

:Tags: Genomics Intervals Summary GFF

Purpose
-------

This script computes the genomic coverage of intervals in
a :term:`gff` formatted file. The coverage is computed per feature.

Usage
-----

You can use two methods to compute the coverage: **genomic** and **histogram**.

Let us explain their usage with this ``small.gtf`` file::

   19 processed_transcript exon 16 16 . - . gene_id
   19 processed_transcript exon 27 27 . - . gene_id
   19 processed_transcript exon 8  8  . - . gene_id
   19 processed_transcript exon 19 19 . - . gene_id
   19 processed_transcript exon 5  5  . - . gene_id

and this toy example (``small.fasta``) of an indexed fasta file::

   >chr19
   GCCGGCCTCTACCTGCAGCAGATGCCCTAT

Both files (``small.gtf`` and ``small.fasta``) are included
in the `GitHub <https://github.com/CGATOxford/cgat>`_ repository.

genomic method
++++++++++++++

The **genomic** method computes the coverage of intervals
accross the genome file given as input. Let us see how to
apply the genomic method to the small examples above::

   python gff2coverage.py --method=genomic --genome-file=small < small.gtf

The output (wrapped to fit here) will be::

   contig  source  feature  intervals  bases  p_coverage  total_p_coverage
   19      trans.  exon     5          5      16.666667   16.666667

As you can see the information displayed is the following: contig name,
source, feature name, number of intervals within the contig, number of
bases, percentage of coverage in the contig, and percentage of coverage
in the genome file.

histogram method
++++++++++++++++

On the contrary, if you want to compute the coverage of intervals
within the :term:`gff` file itself summarized as an histogram and
grouped by contig name, please use the histogram method.

To use the histogram method with the input files above, please type::

 python gff2coverage.py\
 --method=histogram\
 --window=5\
 --features=exon\
 --output-filename-pattern=%s.hist < small.gtf

In this case the output (written to file ``19.hist``) is::

   abs_pos  rel_pos  abs_exon  rel_exon
   0        0.0000   1         0.2000
   5        0.1852   2         0.4000
   10       0.3704   2         0.4000
   15       0.5556   4         0.8000
   20       0.7407   4         0.8000
   25       0.9259   5         1.0000

The output is given as a pair of columns. The first pair of columns always
appears and lists the cumulative numbers of nucleotides in each window or
bin --absolute and relative values in the former and latter columns,
respectively. The subsequent pair of columns depends on the values given to
the ``--features`` option. In this example there is an extra column for the
``exon`` feature but you could especify as many of them as you wanted among
those features listed in your :term:`gff` file.

On the other hand, the ``--num-bins`` option can be used instead of
``--window`` along with ``--genome-file`` to define the number of bins for the
resultant histogram. This parameter is used by default (with value: 1000)
when using the histogram method.

Please note the following:

- you need to specify the feature name explicitly (with the ``--feature`` \
option) to compute the genomic coverage of that feature. You can also use\
a comma-separated list of feature names.

- the output of the histogram method goes to a file (in the current working\
directory) which is named as the contig name by default. To change this\
behaviour, please use the ``--output-filename-pattern`` option where \
``%s`` will be substituted by the contig name.

Command line options
--------------------
'''

import sys
import math
import collections

import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.GTF as GTF


def printValues(contig, max_size, window_size, values, options):
    """output values."""

    outfile = E.openOutputFile(contig, "w")

    outfile.write("abs_pos\trel_pos")

    for feature in options.features:
        outfile.write("\tabs_%s\trel_%s" % (feature, feature))
    outfile.write("\n")

    max_vv = []

    for f in range(len(options.features)):
        max_vv.append(float(max([x[f] for x in values])))

    bin = 0
    for vv in values:
        outfile.write("%i\t" % bin)
        outfile.write(options.value_format % (float(bin) / max_size))

        for x in range(len(options.features)):
            outfile.write("\t%i\t%s" % (
                vv[x],
                options.value_format % (vv[x] / max_vv[x])))
        outfile.write("\n")
        bin += window_size

    outfile.close()


def processChunk(contig, chunk, options, fasta=None):
    """
    This function requires segments to be non-overlapping.
    """

    if len(chunk) == 0:
        return

    # check whether there are overlapping features or not
    checked = []
    for feature in chunk:
        checked.append(feature)
        others = [x for x in chunk if x not in checked]
        for otherFeature in others:
            if GTF.Overlap(feature, otherFeature):
                raise ValueError(" Histogram could not be created"
                                 " since the file contains overlapping "
                                 "features! \n%s\n%s  "
                                 % (feature, otherFeature))
    # clear auxiliary list
    del checked[:]

    # compute max_coordinate for the histogram
    max_coordinate = max([x.end for x in chunk])
    # compute window size
    if options.window_size:
        window_size = options.window_size
        num_bins = int(math.ceil((float(max_coordinate) / window_size)))
    elif options.num_bins and fasta:
        contig_length = fasta.getLength(contig)
        assert max_coordinate <= contig_length, ("maximum coordinate (%i) "
                                                 "larger than contig size (%i)"
                                                 " for contig %s"
                                                 % (max_coordinate,
                                                    contig_length,
                                                    contig))
        max_coordinate = contig_length
        window_size = int(math.floor(float(contig_length) / options.num_bins))
        num_bins = options.num_bins
    else:
        raise ValueError("please specify a window size of provide "
                         "genomic sequence with number of bins.")

    values = [[] for x in range(num_bins)]

    # do several parses for each feature, slow, but easier to code
    # alternatively: sort by feature and location.
    for feature in options.features:
        total = 0
        bin = 0
        end = window_size
        for entry in chunk:
            if entry.feature != feature:
                continue

            while end < entry.start:
                values[bin].append(total)
                bin += 1
                end += window_size

            while entry.end > end:
                seg_start = max(entry.start, end - window_size)
                seg_end = min(entry.end, end)
                total += seg_end - seg_start
                values[bin].append(total)
                end += window_size
                bin += 1
            else:
                seg_start = max(entry.start, end - window_size)
                seg_end = min(entry.end, end)
                total += seg_end - seg_start

        while bin < num_bins:
            values[bin].append(total)
            bin += 1

    printValues(contig, max_coordinate, window_size, values, options)


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(version="%prog version: "
                            "$Id: gff2coverage.py 2781 2009-09-10 11:33:14Z "
                            "andreas $",
                            usage=globals()["__doc__"])

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default]")

    parser.add_option("-f", "--features", dest="features", type="string",
                      action="append", help="features to collect "
                      "[default=%default]")

    parser.add_option("-w", "--window-size", dest="window_size", type="int",
                      help="window size in bp for histogram computation. "
                      "Determines the bin size.  "
                      "[default=%default]")

    parser.add_option("-b", "--num-bins", dest="num_bins", type="int",
                      help="number of bins for histogram computation "
                      "if window size is not given. "
                      "[default=%default]")

    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices=("genomic", "histogram", ),
                      help="methods to apply. "
                      "[default=%default]")

    parser.set_defaults(
        genome_file=None,
        window_size=None,
        num_bins=1000,
        value_format="%6.4f",
        features=[],
        method="genomic",
    )

    (options, args) = E.Start(parser, add_output_options=True)

    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta(options.genome_file)
    else:
        fasta = None

    if options.method == "histogram":

        gff = GTF.readFromFile(options.stdin)

        gff.sort(key=lambda x: (x.contig, x.start))

        chunk = []
        last_contig = None

        for entry in gff:

            if last_contig != entry.contig:
                processChunk(last_contig, chunk, options, fasta)
                last_contig = entry.contig
                chunk = []

            chunk.append(entry)

        processChunk(last_contig, chunk, options, fasta)

    elif options.method == "genomic":
        intervals = collections.defaultdict(int)
        bases = collections.defaultdict(int)
        total = 0
        for entry in GTF.iterator(options.stdin):
            intervals[(entry.contig, entry.source, entry.feature)] += 1
            bases[(entry.contig, entry.source, entry.feature)
                  ] += entry.end - entry.start
            total += entry.end - entry.start

        options.stdout.write("contig\tsource\tfeature\tintervals\tbases")
        if fasta:
            options.stdout.write(
                "\tpercent_coverage\ttotal_percent_coverage\n")
        else:
            options.stdout.write("\n")

        total_genome_size = sum(
            fasta.getContigSizes(with_synonyms=False).values())

        for key in sorted(intervals.keys()):
            nbases = bases[key]
            nintervals = intervals[key]
            contig, source, feature = key
            options.stdout.write("\t".join(("\t".join(key),
                                            str(nintervals),
                                            str(nbases))))
            if fasta:
                options.stdout.write(
                    "\t%f" % (100.0 * float(nbases) / fasta.getLength(contig)))
                options.stdout.write(
                    "\t%f\n" % (100.0 * float(nbases) / total_genome_size))
            else:
                options.stdout.write("\n")

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
MatlabTools.py - 
======================================================

:Tags: Python

Code
----

'''
import sys
import string
import re
import numpy


def WriteMatrixOld(matrix, separator="\t"):
    # this is clumsy

    lines = []
    for x in range(0, matrix.shape[0]):
        lines.append(string.join(list(map(str, matrix[x, ])), separator))

    return string.join(lines, "\n")


def WriteMatrix(matrix, outfile=sys.stdout, separator="\t", format="%f",
                row_headers=None, col_headers=None):
    """write matrix to stream.
    """
    if col_headers:
        outfile.write(separator + separator.join(col_headers) + "\n")

    for x in range(0, matrix.shape[0]):
        if row_headers:
            outfile.write(row_headers[x] + separator)
        outfile.write(
            string.join([format % x for x in matrix[x, ]], separator) + "\n")


def ReadMatrix(file,
               separator="\t",
               numeric_type=numpy.float,
               take="all",
               headers=False
               ):
    """read a matrix. There probably is a routine for this in Numpy, which
    I haven't found yet.
    """

    lines = [x for x in file.readlines() if x[0] != "#"]

    row_headers, col_headers = [], []

    if headers:
        col_headers = lines[0][:-1].split("\t")[1:]
        del lines[0]

    num_rows = len(lines)

    if take != "all":
        num_cols = len(take)
    else:
        l = len(string.split(lines[0][:-1], "\t"))
        if headers:
            take = list(range(1, l))
        else:
            take = list(range(0, l))

    num_cols = len(take)

    matrix = numpy.zeros((num_rows, num_cols), numeric_type)

    nrow = 0
    for l in lines:
        data = l[:-1].split("\t")
        if headers:
            row_headers.append(data[0])

        try:
            data = [float(data[x]) for x in take]
        except ValueError:
            print("error parsing data", data)
            raise

        matrix[nrow] = data
        nrow += 1

    return matrix, row_headers, col_headers


def ReadSparseMatrix(filename,
                     separator="\t",
                     numeric_type=numpy.float,
                     is_symmetric=None):
    """read sparse matrix."""

    lines = [string.split(x[:-1], separator)[:3]
             for x in [x for x in open(filename, "r").readlines() if x[0] != "#"]]

    data = [(string.atoi(x[0]), string.atoi(x[1]), string.atof(x[2]))
            for x in lines]

    num_rows = max([x[0] for x in data])
    num_cols = max([x[1] for x in data])

    if (is_symmetric):
        num_rows = max(num_rows, num_cols)
        num_cols = max(num_rows, num_cols)

    matrix = numpy.zeros((num_rows, num_cols), numeric_type)

    for row, col, weight in data:
        matrix[row - 1, col - 1] = weight

    if is_symmetric:
        for col, row, weight in data:
            matrix[row - 1, col - 1] = weight

    return matrix


def ReadBinarySparseMatrix(filename,
                           separator="\t",
                           numeric_type=numpy.float,
                           is_symmetric=None):
    """read sparse matrix."""

    lines = [string.split(x[:-1], separator)[:2]
             for x in [x for x in open(filename, "r").readlines() if x[0] != "#"]]

    data = [(string.atoi(x[0]), string.atoi(x[1])) for x in lines]

    num_rows = max([x[0] for x in data])
    num_cols = max([x[1] for x in data])

    if (is_symmetric):
        num_rows = max(num_rows, num_cols)
        num_cols = max(num_rows, num_cols)
    matrix = numpy.zeros((num_rows, num_cols), numeric_type)

    for row, col in data:
        matrix[row - 1, col - 1] = 1

    if is_symmetric:
        for col, row in data:
            matrix[row - 1, col - 1] = 1

    return matrix


def readMatrix(infile,
               format="full",
               separator="\t",
               numeric_type=numpy.float,
               take="all",
               headers=True,
               missing=None,
               ):
    """read a matrix from file ane return a numpy matrix.

    formats accepted are:
    * full
    * sparse
    * phylip
    """

    row_headers, col_headers = [], []

    lines = [x for x in infile.readlines() if x[0] != "#"]

    if len(lines) == 0:
        raise IOError("no input")

    if format == "full":

        if headers:
            col_headers = lines[0][:-1].split("\t")[1:]
            del lines[0]

        num_rows = len(lines)

        if take != "all":
            num_cols = len(take)
        else:
            l = len(string.split(lines[0][:-1], "\t"))
            if headers:
                take = list(range(1, l))
            else:
                take = list(range(0, l))

        num_cols = len(take)

        matrix = numpy.zeros((num_rows, num_cols), numeric_type)

        nrow = 0
        for l in lines:
            data = l[:-1].split("\t")
            if headers:
                row_headers.append(data[0])

            if missing is None:
                try:
                    data = [float(data[x]) for x in take]
                except ValueError as msg:
                    raise ValueError("error %s: data=%s" % (msg, str(data)))
                except IndexError as msg:
                    raise IndexError("error %s: data=%s" % (msg, str(data)))

            else:
                d = []
                for x in take:
                    try:
                        d.append(float(data[x]))
                    except ValueError:
                        d.append(missing)
                    except IndexError as msg:
                        raise IndexError(
                            "error %s: data=%s" % (msg, str(data)))

                data = d

            matrix[nrow] = data

            nrow += 1

    elif format == "phylip":
        # read in symmetric phylip matrices
        # note: they can wrap around
        if take != "all":
            raise ValueError(
                "phylip matrix does not support take - "
                "only full matrices are processed.")

        if not headers:
            raise ValueError("phylip matrix always has headers.")

        num_rows = int(lines[0].strip())

        num_cols = num_rows

        matrix = numpy.zeros((num_rows, num_cols), numeric_type)
        take = list(range(1, num_rows))

        nrow = 0
        ncol = 0
        for l in lines[1:]:

            data = re.split("\s+", l[:-1])

            if ncol == 0:
                row_headers.append(data[0])

            try:
                data = list(map(float, data[1:len(data)]))
            except ValueError:
                raise ValueError(
                    "parsing error in conversion to "
                    "float in line %s" % l)

            for x in range(len(data)):
                matrix[nrow][ncol] = data[x]
                ncol += 1

            # deal with wrapping around
            if ncol == num_cols:
                ncol = 0
                nrow += 1

        col_headers = row_headers

    return matrix, row_headers, col_headers


def writeMatrix(outfile, matrix,
                format="full",
                separator="\t",
                value_format="%f",
                row_headers=None,
                col_headers=None):
    """write matrix to stream.
    """

    if format == "full":
        if col_headers:
            outfile.write(separator + separator.join(col_headers) + "\n")

        for x in range(0, matrix.shape[0]):
            if row_headers:
                outfile.write(row_headers[x] + separator)
            outfile.write(
                string.join([value_format % x for x in matrix[x, ]], separator) + "\n")

    elif format == "phylip":
        if not row_headers:
            raise ValueError("phylip output requires row headers")

        nrows = len(row_headers)
        outfile.write("%i\n" % nrows)

        for x in range(0, nrows):
            outfile.write(row_headers[x] + separator)
            outfile.write(
                separator.join([value_format % y for y in matrix[x, ]]) + "\n")

    else:
        raise ValueError("unknown output format %s" % format)



from pysam.libchtslib cimport *
from pysam.libcalignmentfile cimport *

import collections, array, struct, itertools
import CGAT.Experiment as E

cdef class SetNH:

    cdef iter
    cdef stack

    def __cinit__(self, iter ):
        self.iter = itertools.groupby(iter, lambda x: x.qname)
        self.stack = []

    def __iter__(self):
        return self

    def __next__(self): 
        """python version of next().
        """
        
        while 1:
            if self.stack:
                return self.stack.pop(0)
            else:
                key, x = next(self.iter)
                self.stack = list(x)
                nh = len(self.stack)
                for read in self.stack:
                    if not read.is_unmapped:
                        # deal with paired end reads counted
                        # as multi-mapping
                        if read.is_proper_pair and nh > 1:
                            nh -= 1
                        read.set_tag("NH", nh)


def filter_bam(AlignmentFile input_samfile,
               AlignmentFile output_samfile,
               AlignmentFile reference_samfile,
               remove_nonunique = False,
               remove_unique = False,
               remove_contigs = None,
               remove_unmapped = False,
               remove_mismatches = False,
               colour_mismatches = False ):

    '''To conserve memory, the tid and NM flag from *transcripts_samfile*
    are packed into memory. As a consequence, this method requires
    max(NM) < 256 (2^8) and max(tid) < 16777216 (2^24)

    If *remove_nonunique* is set, only uniquely matching reads will be
    output.

    If *remove_contigs* is given, contigs that are in remove_contigs
    will be removed. Note that this will only remove the alignment,
    but not all alignments of a particuluar read and the NH flag will
    *NOT* be updated.

    If *remove_unmapped* is given, unmapped reads will be removed.

    If *remove_mismatches* is set, only reads with number of
    mismatches better than in reference_samfile will be kept.

    If *colour_mismatches* is set, the ``CM`` tag will be used to
    count differences. By default, the ``NM`` tag is used.  The tag
    that is used needs to present in both *input_samfile* and
    *reference_samfile*.

    Detecting non-unique matches:
    
    This method first checks for the NH flag - if set, a unique match
    should have at most NH=1 hits.

    If not set, the method checks for BWA flags. Currently it checks
    if X0 is set (X0=Number of best hits found by BWA). Other relevant
    flags but currently not tested:

    * X1 = Number of suboptimal hits found by BWA
    * XT = Type: Unique/Repeat/N/Mate-sw

    '''

    cdef int ninput = 0
    cdef int nunmapped = 0
    cdef int nmismatches = 0
    cdef int noutput = 0
    cdef int nnonunique = 0
    cdef int nunique = 0
    cdef int nremoved_contigs = 0

    cdef bint c_remove_nonunique = remove_nonunique
    cdef bint c_remove_unique = remove_unique
    cdef bint c_remove_mismatches = remove_mismatches
    cdef bint c_remove_unmapped = remove_unmapped

    cdef int * remove_contig_tids 
    cdef int nremove_contig_tids = 0
    cdef AlignedSegment read
    cdef AlignedSegment match

    # build index
    # this method will start indexing from the current file position
    # if you decide
    cdef int ret = 1
    cdef int x
    cdef bam1_t * b = <bam1_t*>calloc(1, sizeof( bam1_t))
    cdef uint64_t pos
    cdef uint8_t * v
    cdef int32_t nm
    cdef int32_t nh
    cdef int tid
    cdef int transript_tid
    cdef long val
    cdef bint skip = 0

    cdef int32_t read_mismatches = 0

    # decide which tag to use
    cdef char * nm_tag = 'NM' 
    cdef char * cm_tag = 'CM'
    cdef char * tag

    if colour_mismatches:
        tag = cm_tag
    else:
        tag = nm_tag

    if c_remove_mismatches:
        E.info( "building index" )
        if not reference_samfile:
            raise ValueError("require another bam file for mismatch filtering" )

        # L = 1 byte (unsigned char)
        def _gen():
            return array.array('B')
        
        index = collections.defaultdict(_gen)

        while ret > 0:
            ret = bam_read1(
                hts_get_bgzfp(reference_samfile.htsfile),
                b)
            if ret > 0:
                # ignore unmapped reads
                if b.core.flag & 4:
                    continue
                qname = pysam_bam_get_qname(b)
                tid = b.core.tid
                v = bam_aux_get(b, tag)
                if v != NULL:
                    nm = <int32_t>bam_aux2i(v)
                else:
                    nm = 0
                index[qname].append(nm)
        E.info( "built index for %i reads" % len(index))
        bam_destroy1(b)

    # setup list of contigs to remove:
    if remove_contigs:
        nremove_contig_tids = len(remove_contigs)
        remove_contig_tids = <int*>malloc(sizeof(int) * nremove_contig_tids)
        for x, rname in enumerate(remove_contigs):
            remove_contig_tids[x] = input_samfile.gettid(rname)
                
    E.info( "starting filtering" )

    for read in input_samfile:

        ninput += 1
    
        # remove unmapped reads
        if read._delegate.core.flag & 4:
            if c_remove_unmapped:
                nunmapped += 1
            else:
                output_samfile.write( read )
            continue

        # remove non-unique alignments
        if c_remove_nonunique:
            # check either NH or X0 (bwa) flag
            v = bam_aux_get(read._delegate, 'NH')
            if v == NULL:
                v = bam_aux_get(read._delegate, 'X0')
            
            if v != NULL:
                nh = <int32_t>bam_aux2i(v)
                if nh > 1: 
                    nnonunique += 1
                    continue
        # remove unique alignments
        elif c_remove_unique:
            # check either NH or X0 (bwa) flag
            v = bam_aux_get(read._delegate, 'NH')
            if v == NULL:
                v = bam_aux_get(read._delegate, 'X0')
            
            if v != NULL:
                nh = <int32_t>bam_aux2i(v)
                if nh == 1: 
                    nunique += 1
                    continue

        # remove reads matched to certain contigs
        if nremove_contig_tids:
            skip = 0
            for x from 0 <= x < nremove_contig_tids:
                if remove_contig_tids[x] == read.tid:
                    skip = 1
                    break
            if skip: 
                nremoved_contigs += 1
                continue

        # remove reads in other bam file if their are better matching
        if c_remove_mismatches:
            if read.qname in index:
                # can compress index before, depends on
                # how many reads you expect test to filter
                nm = min(index[read.qname])
                read_mismatches = read.opt(tag)
                if nm > read_mismatches: 
                    nmismatches += 1
                    continue

        noutput += 1
        output_samfile.write(read)

    c = E.Counter()
    c.input = ninput
    c.removed_unique = nunique
    c.removed_nonunique = nnonunique
    c.removed_contigs = nremoved_contigs
    c.output = noutput
    c.removed_unmapped = nunmapped
    c.removed_mismatches = nmismatches

    if nremove_contig_tids:
        free(remove_contig_tids)

    E.info("filtering finished")

    return c



'''
SVGdraw.py - generate SVG drawings
======================================================

:Tags: Python

This module has been copied from 3rd party resources.

SVGdraw uses an object model drawing and a method toXML to create SVG graphics
by using easy to use classes and methods usualy you start by creating a drawing eg

    d=drawing()
    #then you create a SVG root element
    s=svg()
    #then you add some elements eg a circle and add it to the svg root element
    c=circle()
    #you can supply attributes by using named arguments.
    c=circle(fill='red',stroke='blue')
    #or by updating the attributes attribute:
    c.attributes['stroke-width']=1
    s.addElement(c)
    #then you add the svg root element to the drawing
    d.setSVG(s)
    #and finaly you xmlify the drawing
    d.toXml()
    

this results in the svg source of the drawing, which consists of a circle
on a white background. Its as easy as that;)
This module was created using the SVG specification of www.w3c.org and the
O'Reilly (www.oreilly.com) python books as information sources. A svg viewer
is available from www.adobe.com
'''

from six import StringIO
import sys

__version__ = "1.0"

use_dom_implementation = 0

if use_dom_implementation != 0:
    try:
        from xml.dom import implementation
        from xml.dom.ext import PrettyPrint
    except:
        raise ImportError("PyXML is required for using the dom implementation")


sys.setrecursionlimit = 50


def _escape(data, entities={}):
    """Escape &, <, and > in a string of data.

    You can escape other strings of data by passing a dictionary as
    the optional entities parameter.  The keys and values must all be
    strings; each key will be replaced with its corresponding value.
    """
    data = data.replace("&", "&amp;")
    data = data.replace("<", "&lt;")
    data = data.replace(">", "&gt;")
    for chars, entity in list(entities.items()):
        data = data.replace(chars, entity)
    return data


def _quoteattr(data, entities={}):
    """Escape and quote an attribute value.

    Escape &, <, and > in a string of data, then quote it for use as
    an attribute value.  The \" character will be escaped as well, if
    necessary.

    You can escape other strings of data by passing a dictionary as
    the optional entities parameter.  The keys and values must all be
    strings; each key will be replaced with its corresponding value.
    """
    data = _escape(data, entities)
    if '"' in data:
        if "'" in data:
            data = '"%s"' % data.replace('"', "&quot;")
        else:
            data = "'%s'" % data
    else:
        data = '"%s"' % data
    return data


def _xypointlist(a):
    """formats a list of xy pairs"""
    s = ''
    for e in a:  # this could be done more elegant
        s += str(e)[1:-1] + '  '
    return s


def _viewboxlist(a):
    """formats a tuple"""
    s = ''
    for e in a:
        s += str(e) + ' '
    return s


def _pointlist(a):
    """formats a list of numbers"""
    return str(a)[1:-1]


class pathdata:

    """class used to create a pathdata object which can be used for a path.
    although most methods are pretty straightforward it might be useful to look at the SVG specification."""
    # I didn't test the methods below.

    def __init__(self, x=None, y=None):
        self.path = []
        if x is not None and y is not None:
            self.path.append('M ' + str(x) + ' ' + str(y))

    def closepath(self):
        """ends the path"""
        self.path.append('z')

    def move(self, x, y):
        """move to absolute"""
        self.path.append('M ' + str(x) + ' ' + str(y))

    def relmove(self, x, y):
        """move to relative"""
        self.path.append('m ' + str(x) + ' ' + str(y))

    def line(self, x, y):
        """line to absolute"""
        self.path.append('L ' + str(x) + ' ' + str(y))

    def relline(self, x, y):
        """line to relative"""
        self.path.append('l ' + str(x) + ' ' + str(y))

    def hline(self, x):
        """horizontal line to absolute"""
        self.path.append('H' + str(x))

    def relhline(self, x):
        """horizontal line to relative"""
        self.path.append('h' + str(x))

    def vline(self, y):
        """verical line to absolute"""
        self.path.append('V' + str(y))

    def relvline(self, y):
        """vertical line to relative"""
        self.path.append('v' + str(y))

    def bezier(self, x1, y1, x2, y2, x, y):
        """bezier with xy1 and xy2 to xy absolut"""
        self.path.append('C' + str(x1) + ',' + str(y1) + ' ' +
                         str(x2) + ',' + str(y2) + ' ' + str(x) + ',' + str(y))

    def relbezier(self, x1, y1, x2, y2, x, y):
        """bezier with xy1 and xy2 to xy relative"""
        self.path.append('c' + str(x1) + ',' + str(y1) + ' ' +
                         str(x2) + ',' + str(y2) + ' ' + str(x) + ',' + str(y))

    def smbezier(self, x2, y2, x, y):
        """smooth bezier with xy2 to xy absolut"""
        self.path.append(
            'S' + str(x2) + ',' + str(y2) + ' ' + str(x) + ',' + str(y))

    def relsmbezier(self, x2, y2, x, y):
        """smooth bezier with xy2 to xy relative"""
        self.path.append(
            's' + str(x2) + ',' + str(y2) + ' ' + str(x) + ',' + str(y))

    def qbezier(self, x1, y1, x, y):
        """quadratic bezier with xy1 to xy absolut"""
        self.path.append(
            'Q' + str(x1) + ',' + str(y1) + ' ' + str(x) + ',' + str(y))

    def relqbezier(self, x1, y1, x, y):
        """quadratic bezier with xy1 to xy relative"""
        self.path.append(
            'q' + str(x1) + ',' + str(y1) + ' ' + str(x) + ',' + str(y))

    def smqbezier(self, x, y):
        """smooth quadratic bezier to xy absolut"""
        self.path.append('T' + str(x) + ',' + str(y))

    def relsmqbezier(self, x, y):
        """smooth quadratic bezier to xy relative"""
        self.path.append('t' + str(x) + ',' + str(y))

    def ellarc(self, rx, ry, xrot, laf, sf, x, y):
        """elliptival arc with rx and ry rotating with xrot using large-arc-flag and sweep-flag  to xy absolut"""
        self.path.append('A' + str(rx) + ',' + str(ry) + ' ' + str(xrot) +
                         ' ' + str(laf) + ' ' + str(sf) + ' ' + str(x) + ' ' + str(y))

    def relellarc(self, rx, ry, xrot, laf, sf, x, y):
        """elliptival arc with rx and ry rotating with xrot using large-arc-flag and sweep-flag  to xy relative"""
        self.path.append('a' + str(rx) + ',' + str(ry) + ' ' + str(xrot) +
                         ' ' + str(laf) + ' ' + str(sf) + ' ' + str(x) + ' ' + str(y))

    def __repr__(self):
        return ' '.join(self.path)


class SVGelement:

    """SVGelement(type,attributes,elements,text,namespace,**args)
    Creates a arbitrary svg element and is intended to be subclassed not used on its own.
    This element is the base of every svg element it defines a class which resembles
    a xml-element. The main advantage of this kind of implementation is that you don't
    have to create a toXML method for every different graph object. Every element
    consists of a type, attribute, optional subelements, optional text and an optional
    namespace. Note the elements==None, if elements = None:self.elements=[] construction.
    This is done because if you default to elements=[] every object has a reference
    to the same empty list."""

    def __init__(self, type='', attributes=None, elements=None, text='', namespace='', cdata=None, **args):
        self.type = type
        if attributes is None:
            self.attributes = {}
        else:
            self.attributes = attributes
        if elements is None:
            self.elements = []
        else:
            self.elements = elements
        self.text = text
        self.namespace = namespace
        self.cdata = cdata
        for arg in list(args.keys()):
            self.attributes[arg] = args[arg]

    def addElement(self, SVGelement):
        """adds an element to a SVGelement

        SVGelement.addElement(SVGelement)
        """
        self.elements.append(SVGelement)

    def toXml(self, level, f):
        f.write('\t' * level)
        f.write('<' + self.type)
        for attkey in list(self.attributes.keys()):
            f.write(' ' + _escape(str(attkey)) + '=' +
                    _quoteattr(str(self.attributes[attkey])))
        if self.namespace:
            f.write(' xmlns="' + _escape(str(self.namespace)) + '" ')
            # added by Andreas Heger
            f.write(
                ' xmlns:xlink="' + _escape(str("http://www.w3.org/1999/xlink")) + '" ')
        if self.elements or self.text or self.cdata:
            f.write('>')
        if self.elements:
            f.write('\n')
        for element in self.elements:
            element.toXml(level + 1, f)
        if self.cdata:
            f.write('\n' + '\t' * (level + 1) + '<![CDATA[')
            for line in self.cdata.splitlines():
                f.write('\n' + '\t' * (level + 2) + line)
            f.write('\n' + '\t' * (level + 1) + ']]>\n')
        if self.text:
            if isinstance(self.text, str):  # If the text is only text
                f.write(_escape(str(self.text)))
            else:  # If the text is a spannedtext class
                f.write(str(self.text))
        if self.elements:
            f.write('\t' * level + '</' + self.type + '>\n')
        elif self.text:
            f.write('</' + self.type + '>\n')
        elif self.cdata:
            f.write('\t' * level + '</' + self.type + '>\n')
        else:
            f.write('/>\n')


class tspan(SVGelement):

    """ts=tspan(text='',**args)

    a tspan element can be used for applying formatting to a textsection
    usage:
    ts=tspan('this text is bold')
    ts.attributes['font-weight']='bold'
    st=spannedtext()
    st.addtspan(ts)
    t=text(3,5,st)
    """

    def __init__(self, text=None, **args):
        SVGelement.__init__(self, 'tspan', **args)
        if self.text is not None:
            self.text = text

    def __repr__(self):
        s = "<tspan"
        for key, value in list(self.attributes.items()):
            s += ' %s="%s"' % (key, value)
        s += '>'
        s += self.text
        s += '</tspan>'
        return s


class tref(SVGelement):

    """tr=tref(link='',**args)

    a tref element can be used for referencing text by a link to its id.
    usage:
    tr=tref('#linktotext')
    st=spannedtext()
    st.addtref(tr)
    t=text(3,5,st)
    """

    def __init__(self, link, **args):
        SVGelement.__init__(self, 'tref', {'xlink:href': link}, **args)

    def __repr__(self):
        s = "<tref"

        for key, value in list(self.attributes.items()):
            s += ' %s="%s"' % (key, value)
        s += '/>'
        return s


class spannedtext:

    """st=spannedtext(textlist=[])

    a spannedtext can be used for text which consists of text, tspan's and tref's
    You can use it to add to a text element or path element. Don't add it directly
    to a svg or a group element.
    usage:

    ts=tspan('this text is bold')
    ts.attributes['font-weight']='bold'
    tr=tref('#linktotext')
    tr.attributes['fill']='red'
    st=spannedtext()
    st.addtspan(ts)
    st.addtref(tr)
    st.addtext('This text is not bold')
    t=text(3,5,st)
    """

    def __init__(self, textlist=None):
        if textlist is None:
            self.textlist = []
        else:
            self.textlist = textlist

    def addtext(self, text=''):
        self.textlist.append(text)

    def addtspan(self, tspan):
        self.textlist.append(tspan)

    def addtref(self, tref):
        self.textlist.append(tref)

    def __repr__(self):
        s = ""
        for element in self.textlist:
            s += str(element)
        return s


class rect(SVGelement):

    """r=rect(width,height,x,y,fill,stroke,stroke_width,**args)

    a rectangle is defined by a width and height and a xy pair 
    """

    def __init__(self, x=None, y=None, width=None, height=None, fill=None, stroke=None, stroke_width=None, **args):
        if width is None or height is None:
            if width is not None:
                raise ValueError('height is required')
            if height is not None:
                raise ValueError('width is required')
            else:
                raise ValueError('both height and width are required')
        SVGelement.__init__(
            self, 'rect', {'width': width, 'height': height}, **args)
        if x is not None:
            self.attributes['x'] = x
        if y is not None:
            self.attributes['y'] = y
        if fill is not None:
            self.attributes['fill'] = fill
        if stroke is not None:
            self.attributes['stroke'] = stroke
        if stroke_width is not None:
            self.attributes['stroke-width'] = stroke_width


class ellipse(SVGelement):

    """e=ellipse(rx,ry,x,y,fill,stroke,stroke_width,**args)

    an ellipse is defined as a center and a x and y radius.
    """

    def __init__(self, cx=None, cy=None, rx=None, ry=None, fill=None, stroke=None, stroke_width=None, **args):
        if rx is None or ry is None:
            if rx is not None:
                raise ValueError('rx is required')
            if ry is not None:
                raise ValueError('ry is required')
            else:
                raise ValueError('both rx and ry are required')
        SVGelement.__init__(self, 'ellipse', {'rx': rx, 'ry': ry}, **args)
        if cx is not None:
            self.attributes['cx'] = cx
        if cy is not None:
            self.attributes['cy'] = cy
        if fill is not None:
            self.attributes['fill'] = fill
        if stroke is not None:
            self.attributes['stroke'] = stroke
        if stroke_width is not None:
            self.attributes['stroke-width'] = stroke_width


class circle(SVGelement):

    """c=circle(x,y,radius,fill,stroke,stroke_width,**args)

    The circle creates an element using a x, y and radius values eg
    """

    def __init__(self, cx=None, cy=None, r=None, fill=None, stroke=None, stroke_width=None, **args):
        if r is None:
            raise ValueError('r is required')
        SVGelement.__init__(self, 'circle', {'r': r}, **args)
        if cx is not None:
            self.attributes['cx'] = cx
        if cy is not None:
            self.attributes['cy'] = cy
        if fill is not None:
            self.attributes['fill'] = fill
        if stroke is not None:
            self.attributes['stroke'] = stroke
        if stroke_width is not None:
            self.attributes['stroke-width'] = stroke_width


class point(circle):

    """p=point(x,y,color)

    A point is defined as a circle with a size 1 radius. It may be more efficient to use a
    very small rectangle if you use many points because a circle is difficult to render.
    """

    def __init__(self, x, y, fill='black', **args):
        circle.__init__(self, x, y, 1, fill, **args)


class line(SVGelement):

    """l=line(x1,y1,x2,y2,stroke,stroke_width,**args)

    A line is defined by a begin x,y pair and an end x,y pair
    """

    def __init__(self, x1=None, y1=None, x2=None, y2=None, stroke=None, stroke_width=None, **args):
        SVGelement.__init__(self, 'line', **args)
        if x1 is not None:
            self.attributes['x1'] = x1
        if y1 is not None:
            self.attributes['y1'] = y1
        if x2 is not None:
            self.attributes['x2'] = x2
        if y2 is not None:
            self.attributes['y2'] = y2
        if stroke_width is not None:
            self.attributes['stroke-width'] = stroke_width
        if stroke is not None:
            self.attributes['stroke'] = stroke


class polyline(SVGelement):

    """pl=polyline([[x1,y1],[x2,y2],...],fill,stroke,stroke_width,**args)

    a polyline is defined by a list of xy pairs
    """

    def __init__(self, points, fill=None, stroke=None, stroke_width=None, **args):
        SVGelement.__init__(
            self, 'polyline', {'points': _xypointlist(points)}, **args)
        if fill is not None:
            self.attributes['fill'] = fill
        if stroke_width is not None:
            self.attributes['stroke-width'] = stroke_width
        if stroke is not None:
            self.attributes['stroke'] = stroke


class polygon(SVGelement):

    """pl=polyline([[x1,y1],[x2,y2],...],fill,stroke,stroke_width,**args)

    a polygon is defined by a list of xy pairs
    """

    def __init__(self, points, fill=None, stroke=None, stroke_width=None, **args):
        SVGelement.__init__(
            self, 'polygon', {'points': _xypointlist(points)}, **args)
        if fill is not None:
            self.attributes['fill'] = fill
        if stroke_width is not None:
            self.attributes['stroke-width'] = stroke_width
        if stroke is not None:
            self.attributes['stroke'] = stroke


class path(SVGelement):

    """p=path(path,fill,stroke,stroke_width,**args)

    a path is defined by a path object and optional width, stroke and fillcolor
    """

    def __init__(self, pathdata, fill=None, stroke=None, stroke_width=None, id=None, **args):
        SVGelement.__init__(self, 'path', {'d': str(pathdata)}, **args)
        if stroke is not None:
            self.attributes['stroke'] = stroke
        if fill is not None:
            self.attributes['fill'] = fill
        if stroke_width is not None:
            self.attributes['stroke-width'] = stroke_width
        if id is not None:
            self.attributes['id'] = id


class text(SVGelement):

    """t=text(x,y,text,font_size,font_family,**args)

    a text element can bge used for displaying text on the screen
    """

    def __init__(self, x=None, y=None, text=None, font_size=None, font_family=None, text_anchor=None, font_style=None, **args):
        SVGelement.__init__(self, 'text', **args)
        if x is not None:
            self.attributes['x'] = x
        if y is not None:
            self.attributes['y'] = y
        if font_size is not None:
            self.attributes['font-size'] = font_size
        if font_family is not None:
            self.attributes['font-family'] = font_family
        if font_style is not None:
            self.attributes['font-style'] = font_style
        if text is not None:
            self.text = text
        if text_anchor is not None:
            self.attributes['text-anchor'] = text_anchor


class textpath(SVGelement):

    """tp=textpath(text,link,**args)

    a textpath places a text on a path which is referenced by a link.   
    """

    def __init__(self, link, text=None, **args):
        SVGelement.__init__(self, 'textPath', {'xlink:href': link}, **args)
        if text is not None:
            self.text = text


class pattern(SVGelement):

    """p=pattern(x,y,width,height,patternUnits,**args)

    A pattern is used to fill or stroke an object using a pre-defined
    graphic object which can be replicated ("tiled") at fixed intervals
    in x and y to cover the areas to be painted.
    """

    def __init__(self, x=None, y=None, width=None, height=None, patternUnits=None, **args):
        SVGelement.__init__(self, 'pattern', **args)
        if x is not None:
            self.attributes['x'] = x
        if y is not None:
            self.attributes['y'] = y
        if width is not None:
            self.attributes['width'] = width
        if height is not None:
            self.attributes['height'] = height
        if patternUnits is not None:
            self.attributes['patternUnits'] = patternUnits


class title(SVGelement):

    """t=title(text,**args)

    a title is a text element. The text is displayed in the title bar
    add at least one to the root svg element
    """

    def __init__(self, text=None, **args):
        SVGelement.__init__(self, 'title', **args)
        if text is not None:
            self.text = text


class description(SVGelement):

    """d=description(text,**args)

    a description can be added to any element and is used for a tooltip
    Add this element before adding other elements.
    """

    def __init__(self, text=None, **args):
        SVGelement.__init__(self, 'desc', **args)
        if text is not None:
            self.text = text


class lineargradient(SVGelement):

    """lg=lineargradient(x1,y1,x2,y2,id,**args)

    defines a lineargradient using two xy pairs.
    stop elements van be added to define the gradient colors.
    """

    def __init__(self, x1=None, y1=None, x2=None, y2=None, id=None, **args):
        SVGelement.__init__(self, 'linearGradient', **args)
        if x1 is not None:
            self.attributes['x1'] = x1
        if y1 is not None:
            self.attributes['y1'] = y1
        if x2 is not None:
            self.attributes['x2'] = x2
        if y2 is not None:
            self.attributes['y2'] = y2
        if id is not None:
            self.attributes['id'] = id


class radialgradient(SVGelement):

    """rg=radialgradient(cx,cy,r,fx,fy,id,**args)

    defines a radial gradient using a outer circle which are defined by a cx,cy and r and by using a focalpoint.
    stop elements van be added to define the gradient colors.
    """

    def __init__(self, cx=None, cy=None, r=None, fx=None, fy=None, id=None, **args):
        SVGelement.__init__(self, 'radialGradient', **args)
        if cx is not None:
            self.attributes['cx'] = cx
        if cy is not None:
            self.attributes['cy'] = cy
        if r is not None:
            self.attributes['r'] = r
        if fx is not None:
            self.attributes['fx'] = fx
        if fy is not None:
            self.attributes['fy'] = fy
        if id is not None:
            self.attributes['id'] = id


class stop(SVGelement):

    """st=stop(offset,stop_color,**args)

    Puts a stop color at the specified radius
    """

    def __init__(self, offset, stop_color=None, **args):
        SVGelement.__init__(self, 'stop', {'offset': offset}, **args)
        if stop_color is not None:
            self.attributes['stop-color'] = stop_color


class style(SVGelement):

    """st=style(type,cdata=None,**args)

    Add a CDATA element to this element for defing in line stylesheets etc..
    """

    def __init__(self, type, cdata=None, **args):
        SVGelement.__init__(self, 'style', {'type': type}, cdata=cdata, **args)


class image(SVGelement):

    """im=image(url,width,height,x,y,**args)

    adds an image to the drawing. Supported formats are .png, .jpg and .svg.
    """

    def __init__(self, url, x=None, y=None, width=None, height=None, **args):
        if width is None or height is None:
            if width is not None:
                raise ValueError('height is required')
            if height is not None:
                raise ValueError('width is required')
            else:
                raise ValueError('both height and width are required')
        SVGelement.__init__(
            self, 'image', {'xlink:href': url, 'width': width, 'height': height}, **args)
        if x is not None:
            self.attributes['x'] = x
        if y is not None:
            self.attributes['y'] = y


class cursor(SVGelement):

    """c=cursor(url,**args)

    defines a custom cursor for a element or a drawing
    """

    def __init__(self, url, **args):
        SVGelement.__init__(self, 'cursor', {'xlink:href': url}, **args)


class marker(SVGelement):

    """m=marker(id,viewbox,refX,refY,markerWidth,markerHeight,**args)

    defines a marker which can be used as an endpoint for a line or other pathtypes
    add an element to it which should be used as a marker.
    """

    def __init__(self, id=None, viewBox=None, refx=None, refy=None, markerWidth=None, markerHeight=None, **args):
        SVGelement.__init__(self, 'marker', **args)
        if id is not None:
            self.attributes['id'] = id
        if viewBox is not None:
            self.attributes['viewBox'] = _viewboxlist(viewBox)
        if refx is not None:
            self.attributes['refX'] = refx
        if refy is not None:
            self.attributes['refY'] = refy
        if markerWidth is not None:
            self.attributes['markerWidth'] = markerWidth
        if markerHeight is not None:
            self.attributes['markerHeight'] = markerHeight


class group(SVGelement):

    """g=group(id,**args)

    a group is defined by an id and is used to contain elements
    g.addElement(SVGelement)
    """

    def __init__(self, id=None, **args):
        SVGelement.__init__(self, 'g', **args)
        if id is not None:
            self.attributes['id'] = id


class symbol(SVGelement):

    """sy=symbol(id,viewbox,**args)

    defines a symbol which can be used on different places in your graph using
    the use element. A symbol is not rendered but you can use 'use' elements to
    display it by referencing its id.
    sy.addElement(SVGelement)
    """

    def __init__(self, id=None, viewBox=None, **args):
        SVGelement.__init__(self, 'symbol', **args)
        if id is not None:
            self.attributes['id'] = id
        if viewBox is not None:
            self.attributes['viewBox'] = _viewboxlist(viewBox)


class defs(SVGelement):

    """d=defs(``**args``)

    container for defining elements
    """

    def __init__(self, **args):
        SVGelement.__init__(self, 'defs', **args)


class switch(SVGelement):

    """sw=switch(``**args``)

    Elements added to a switch element which are "switched" by the attributes
    requiredFeatures, requiredExtensions and systemLanguage.
    Refer to the SVG specification for details.
    """

    def __init__(self, **args):
        SVGelement.__init__(self, 'switch', **args)


class use(SVGelement):

    """u=use(link,x,y,width,height,``**args``)

    references a symbol by linking to its id and its position, height and width
    """

    def __init__(self, link, x=None, y=None, width=None, height=None, **args):
        SVGelement.__init__(self, 'use', {'xlink:href': link}, **args)
        if x is not None:
            self.attributes['x'] = x
        if y is not None:
            self.attributes['y'] = y

        if width is not None:
            self.attributes['width'] = width
        if height is not None:
            self.attributes['height'] = height


class link(SVGelement):

    """a=link(url,``**args``)

    a link  is defined by a hyperlink. add elements which have to be linked
    a.addElement(SVGelement)
    """

    def __init__(self, link='', **args):
        SVGelement.__init__(self, 'a', {'xlink:href': link}, **args)


class view(SVGelement):

    """v=view(id,``**args``)

    a view can be used to create a view with different attributes"""

    def __init__(self, id=None, **args):
        SVGelement.__init__(self, 'view', **args)
        if id is not None:
            self.attributes['id'] = id


class script(SVGelement):

    """sc=script(type,type,cdata,``**args``)

    adds a script element which contains CDATA to the SVG drawing

    """

    def __init__(self, type, cdata=None, **args):
        SVGelement.__init__(
            self, 'script', {'type': type}, cdata=cdata, **args)


class animate(SVGelement):

    """an=animate(attribute,from,to,during,``**args``)

    animates an attribute.    
    """

    def __init__(self, attribute, fr=None, to=None, dur=None, **args):
        SVGelement.__init__(
            self, 'animate', {'attributeName': attribute}, **args)
        if fr is not None:
            self.attributes['from'] = fr
        if to is not None:
            self.attributes['to'] = to
        if dur is not None:
            self.attributes['dur'] = dur


class animateMotion(SVGelement):

    """an=animateMotion(pathdata,dur,``**args``)

    animates a SVGelement over the given path in dur seconds
    """

    def __init__(self, pathdata, dur, **args):
        SVGelement.__init__(self, 'animateMotion', **args)
        if pathdata is not None:
            self.attributes['path'] = str(pathdata)
        if dur is not None:
            self.attributes['dur'] = dur


class animateTransform(SVGelement):

    """antr=animateTransform(type,from,to,dur,``**args``)

    transform an element from and to a value.
    """

    def __init__(self, type=None, fr=None, to=None, dur=None, **args):
        SVGelement.__init__(
            self, 'animateTransform', {'attributeName': 'transform'}, **args)
        # As far as I know the attributeName is always transform
        if type is not None:
            self.attributes['type'] = type
        if fr is not None:
            self.attributes['from'] = fr
        if to is not None:
            self.attributes['to'] = to
        if dur is not None:
            self.attributes['dur'] = dur


class animateColor(SVGelement):

    """ac=animateColor(attribute,type,from,to,dur,``**args``)

    Animates the color of a element
    """

    def __init__(self, attribute, type=None, fr=None, to=None, dur=None, **args):
        SVGelement.__init__(
            self, 'animateColor', {'attributeName': attribute}, **args)
        if type is not None:
            self.attributes['type'] = type
        if fr is not None:
            self.attributes['from'] = fr
        if to is not None:
            self.attributes['to'] = to
        if dur is not None:
            self.attributes['dur'] = dur


class set(SVGelement):

    """st=set(attribute,to,during,``**args``)

    sets an attribute to a value for a
    """

    def __init__(self, attribute, to=None, dur=None, **args):
        SVGelement.__init__(self, 'set', {'attributeName': attribute}, **args)
        if to is not None:
            self.attributes['to'] = to
        if dur is not None:
            self.attributes['dur'] = dur


class svg(SVGelement):

    """s=svg(viewbox,width,height,``**args``)

    a svg or element is the root of a drawing add all elements to a svg element.
    You can have different svg elements in one svg file
    s.addElement(SVGelement)

    eg
    d=drawing()
    s=svg((0,0,100,100),'100%','100%')
    c=circle(50,50,20)
    s.addElement(c)
    d.setSVG(s)
    d.toXml()
    """

    def __init__(self, viewBox=None, width=None, height=None, **args):
        SVGelement.__init__(self, 'svg', **args)
        if viewBox is not None:
            self.attributes['viewBox'] = _viewboxlist(viewBox)
        if width is not None:
            self.attributes['width'] = width
        if height is not None:
            self.attributes['height'] = height
        self.namespace = "http://www.w3.org/2000/svg"


class drawing:

    """d=drawing()

    this is the actual SVG document. It needs a svg element as a root.
    Use the addSVG method to set the svg to the root. Use the toXml method to write the SVG
    source to the screen or to a file
    d=drawing()
    d.addSVG(svg)
    d.toXml(optionalfilename)
    """

    def __init__(self):
        self.svg = None

    def setSVG(self, svg):
        self.svg = svg
        # Voeg een element toe aan de grafiek toe.
    if use_dom_implementation == 0:
        def toXml(self, filename='', compress=False):
            xml = StringIO()
            xml.write("<?xml version='1.0' encoding='UTF-8'?>\n")
            xml.write(
                "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.0//EN\" \"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd \">\n")
            self.svg.toXml(0, xml)
            if not filename:
                if compress:
                    import gzip
                    f = StringIO()
                    zf = gzip.GzipFile(fileobj=f, mode='wb')
                    zf.write(xml.getvalue())
                    zf.close()
                    f.seek(0)
                    return f.read()
                else:
                    return xml.getvalue()
            else:
                if filename[-4:] == 'svgz':
                    import gzip
                    f = gzip.GzipFile(
                        filename=filename, mode="wb", compresslevel=9)
                    f.write(xml.getvalue())
                    f.close()
                else:
                    f = file(filename, 'w')
                    f.write(xml.getvalue())
                    f.close()

    else:
        def toXml(self, filename='', compress=False):
            """drawing.toXml()        ---->to the screen
            drawing.toXml(filename)---->to the file
            writes a svg drawing to the screen or to a file
            compresses if filename ends with svgz or if compress is true
            """
            doctype = implementation.createDocumentType(
                'svg', "-//W3C//DTD SVG 1.0//EN""", 'http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd ')

            global root
            # root is defined global so it can be used by the appender. Its also possible to use it as an arugument but
            # that is a bit messy.
            root = implementation.createDocument(None, None, doctype)
            # Create the xml document.
            global appender

            def appender(element, elementroot):
                """This recursive function appends elements to an element and sets the attributes
                and type. It stops when alle elements have been appended"""
                if element.namespace:
                    e = root.createElementNS(element.namespace, element.type)
                else:
                    e = root.createElement(element.type)
                if element.text:
                    textnode = root.createTextNode(element.text)
                    e.appendChild(textnode)
                # in element.attributes is supported from python 2.2
                for attribute in list(element.attributes.keys()):
                    e.setAttribute(
                        attribute, str(element.attributes[attribute]))
                if element.elements:
                    for el in element.elements:
                        e = appender(el, e)
                elementroot.appendChild(e)
                return elementroot
            root = appender(self.svg, root)
            if not filename:
                xml = StringIO()
                PrettyPrint(root, xml)
                if compress:
                    import gzip
                    f = StringIO()
                    zf = gzip.GzipFile(fileobj=f, mode='wb')
                    zf.write(xml.getvalue())
                    zf.close()
                    f.seek(0)
                    return f.read()
                else:
                    return xml.getvalue()
            else:
                try:
                    if filename[-4:] == 'svgz':
                        import gzip
                        xml = StringIO()
                        PrettyPrint(root, xml)
                        f = gzip.GzipFile(
                            filename=filename, mode='wb', compresslevel=9)
                        f.write(xml.getvalue())
                        f.close()
                    else:
                        f = open(filename, 'w')
                        PrettyPrint(root, f)
                        f.close()
                except:
                    print("Cannot write SVG file: " + filename)

    def validate(self):
        try:
            import xml.parsers.xmlproc.xmlval
        except:
            raise ImportError('PyXml is required for validating SVG')
        svg = self.toXml()
        xv = xml.parsers.xmlproc.xmlval.XMLValidator()
        try:
            xv.feed(svg)
        except:
            raise ValueError("SVG is not well formed, see messages above")
        else:
            print("SVG well formed")


if __name__ == '__main__':

    d = drawing()
    s = svg((0, 0, 100, 100))
    r = rect(-100, -100, 300, 300, 'cyan')
    s.addElement(r)

    t = title('SVGdraw Demo')
    s.addElement(t)
    g = group('animations')
    e = ellipse(0, 0, 5, 2)
    g.addElement(e)
    c = circle(0, 0, 1, 'red')
    g.addElement(c)
    pd = pathdata(0, -10)
    for i in range(6):
        pd.relsmbezier(10, 5, 0, 10)
        pd.relsmbezier(-10, 5, 0, 10)
    an = animateMotion(pd, 10)
    an.attributes['rotate'] = 'auto-reverse'
    an.attributes['repeatCount'] = "indefinite"
    g.addElement(an)
    s.addElement(g)
    for i in range(20, 120, 20):
        u = use('#animations', i, 0)
        s.addElement(u)
    for i in range(0, 120, 20):
        for j in range(5, 105, 10):
            c = circle(i, j, 1, 'red', 'black', .5)
            s.addElement(c)
    d.setSVG(s)

    print(d.toXml())



"""MAST.py - Parser for MAST files
==================================

:Tags: Python

As of biopython 1.5.6, the MAST parser is broken.
"""

import re
import math
import collections

import numpy


class MAST:
    # number of matches

    def __init__(self):
        self.motifs = []
        self.matches = []


class Motif:
    pass


class Match:

    """a MAST entry.
    """
    header = "\t".join(("id", "description", "length", "pvalue", "evalue",
                        "nmatches", "diagram", "start", "end", "positions", "motifs", "directions"))

    def __init__(self):
        pass

    def __str__(self):
        return "\t".join(map(str,
                             (self.id,
                              self.description,
                              self.length,
                              self.pvalue,
                              self.evalue,
                              self.nmotifs,
                              self.diagram,
                              self.start,
                              self.end,
                              ",".join(map(str, self.positions)),
                              ",".join(map(str, self.motifs)),
                              ",".join(self.directions))))


def parse(infile):
    '''parse verbose MAST output.'''

    line = None
    for line in infile:
        if line.startswith("DATABASE AND MOTIFS"):
            break

    mast = MAST()
    # empty input
    if not line:
        return mast

    keep = False
    for line in infile:
        l = line.strip()
        if l.startswith("MOTIF WIDTH"):
            keep = True
            continue
        if not keep:
            continue
        if keep:
            if l.startswith("-"):
                continue
            if not l:
                break
            motif = Motif()
            motif.id, motif.length, motif.sequence = re.split("\s+", l)
            motif.length = int(motif.length)
            mast.motifs.append(motif)

    for line in infile:
        if line.startswith("SECTION III"):
            break

    assert line, "could not find tag 'SECTION III'"

    def __myiter(infile):
        keep = False
        do_yield = False
        for line in infile:
            if line.startswith("*"):
                continue
            if keep:
                l = line[:-1].strip()
                if do_yield:
                    # collect continuation of DIAGRAM line
                    if len(line) > 2:
                        block[-1] += l
                        continue
                    else:
                        yield block
                        keep = False
                        do_yield = False

                block.append(l)
                # start end of block (DIAGRAM might be multiline)
                if l.startswith("DIAGRAM"):
                    do_yield = True
            else:
                if not re.match("[\d\s]", line):
                    # start of new block
                    keep = True
                    block = [line[:-1].strip()]

    for block in __myiter(infile):
        assert len(block) == 4, "invalid block: %s" % "".join(block)
        match = Match()
        match.id = block[0]
        match.description = block[1]
        match.length, match.pvalue, match.evalue = re.match(
            "LENGTH =\s*(\d+)\s*COMBINED P-VALUE =\s*(\S+)\s*E-VALUE =\s*(\S+)", block[2]).groups()
        match.length, match.pvalue, match.evalue = \
            int(match.length), float(match.pvalue), float(match.evalue)
        match.diagram = re.match("DIAGRAM:\s+(\S+)", block[3]).groups()[0]
        parts = match.diagram.split("_")
        match.nmotifs = len(parts) // 2
        match.positions, match.directions, match.motifs = [], [], []
        match.start = ""
        match.end = ""
        if match.nmotifs > 0:
            pos = 0
            last_end = 0
            try:
                for part in parts:
                    is_motif = part[0] in "[<"
                    if is_motif:
                        code = part[1:-1]
                        match.positions.append(pos)
                        match.directions.append(code[0])
                        motif_id = int(code[1:])
                        match.motifs.append(motif_id)
                        pos += mast.motifs[motif_id - 1].length
                        last_end = pos
                    else:
                        pos += int(part)
            except ValueError:
                raise ValueError(
                    "ParsingError for motif diagram %s" % match.diagram)

            match.start = match.positions[0]
            assert last_end > 0
            match.end = last_end

        mast.matches.append(match)

    return mast


def frequencies2logodds(counts, background_frequencies=None):
    '''write a motif from *counts* to outfile.

    Counts should be a numpy matrix with *nalphabet* columns and *motif_width* rows.
    '''

    motif_width, nalphabet = counts.shape

    if background_frequencies is None:
        # use uniform background
        f = 1.0 / nalphabet
        background_frequencies = [f] * nalphabet

    assert len(background_frequencies) == nalphabet
    background_freqs = numpy.array(background_frequencies)
    logodds_matrix = numpy.ones((motif_width, nalphabet), dtype=numpy.float)

    for x in range(motif_width):
        # divide by backgound_freqs
        # divide by column total
        # take log
        logodds_matrix[x] = numpy.log(
            counts[x] / counts[x].sum() / background_freqs)

    return logodds_matrix


def writeMast(outfile, logodds_matrix, alphabet):
    '''output logodds matrix in MAST format.'''

    motif_width, nalphabet, = logodds_matrix.shape
    assert len(alphabet) == nalphabet

    outfile.write("ALPHABET= %s\n" % alphabet)
    outfile.write("log-odds matrix: alength= %i w= %i\n" %
                  (nalphabet, motif_width))
    for row in logodds_matrix:
        outfile.write(" ".join(
            ["%5.3f" % x for x in row]) + "\n")

    outfile.write("\n")


def writeTomTom(outfile, counts_matrix, header=False):
    '''output counts matrix in tomtom format.

    output counts with columns as motif positions
    and rows as alphabet.
    '''

    if header:
        outfile.write(
            '--------------------------------------------------------------------------------\n')
        outfile.write('        Motif 1 position-specific probability matrix\n')
        outfile.write(
            '--------------------------------------------------------------------------------\n')
        outfile.write(
            "letter-probability matrix: alength= %i w= %i nsites= 18 E= 1.0\n" % counts_matrix.shape)

    for row in numpy.transpose(counts_matrix):
        outfile.write(" ".join(
            ["%6i" % x for x in row]) + "\n")

    outfile.write("\n")


def sequences2motif(outfile, sequences, background_frequencies=None, format="MAST"):
    '''write a motif defined by a collection of sequences to outfile.'''

    # collect all letters
    counts = collections.defaultdict(int)
    nsequences = len(sequences)
    seqs = [re.sub("\s", "", x.upper()) for x in sequences]
    for s in seqs:
        for c in s:
            counts[c] += 1

    lengths = [len(x) for x in seqs]
    motif_width = min(lengths)
    assert max(lengths) == motif_width, "seqs have to have the same length"
    alphabet = "".join(sorted(counts.keys()))
    lalphabet = len(alphabet)
    char2index = dict([(y, x) for x, y in enumerate(alphabet)])

    if background_frequencies is None:
        # use uniform background
        f = 1.0 / lalphabet
        background_frequencies = [f] * lalphabet

    if format == "MAST":
        outfile.write("ALPHABET= %s\n" % alphabet)
        outfile.write("log-odds matrix: alength= %i w= %i\n" %
                      (lalphabet, motif_width))
        for r in range(0, motif_width):
            # use a pseudocount of 1
            row_counts = [1] * lalphabet
            for s in seqs:
                row_counts[char2index[s[r]]] += 1
            outfile.write(" ".join(
                ["%5.3f" % (math.log(float(row_counts[x]) / nsequences / background_frequencies[x]))
                 for x in range(lalphabet)]) + "\n")
    elif format == "probability":
        for r in range(0, motif_width):
            # use a pseudocount of 1
            row_counts = [1] * lalphabet
            for s in seqs:
                row_counts[char2index[s[r]]] += 1
            outfile.write(" ".join(
                ["%5.3f" % (float(row_counts[x]) / nsequences)
                 for x in range(lalphabet)]) + "\n")

    elif format == "TOMTOM":
        counts = numpy.zeros((motif_width, lalphabet))
        for x in range(0, motif_width):
            for s in seqs:
                counts[x, char2index[s[x]]] += 1

        for y in range(lalphabet):
            for x in range(motif_width):
                outfile.write(" %6i" % counts[x, y])
            outfile.write("\n")



'''
Motifs.py - 
======================================================

:Tags: Python

Code
----

'''
import collections
from CGAT import Genomics as Genomics
from CGAT import FastaIterator as FastaIterator


def countMotifs(infile, motifs):
    '''find regular expression *motifs* in
    sequences within fasta formatted *infile*.
    '''

    it = FastaIterator.FastaIterator(infile)
    positions = []
    while 1:
        try:
            seq = next(it)
        except StopIteration:
            break
        if not seq:
            break

        rseq = Genomics.complement(seq.sequence)
        lsequence = len(seq.sequence)
        pos = []
        for motif, pattern in motifs:

            for x in pattern.finditer(seq.sequence):
                pos.append((motif, "+", x.start(), x.end()))
            for x in pattern.finditer(rseq):
                pos.append(
                    (motif, "-", lsequence - x.end(), lsequence - x.start()))

        positions.append((seq.title, pos))

    return positions


def getCounts(matches):
    '''count numbers of motifs.'''
    counts = collections.defaultdict(int)
    for seq, pos in matches:
        for motif, strand, start, end in pos:
            counts[motif] += 1
    return counts


def getOccurances(matches):
    '''count numbers of motifs, but only once per sequence'''
    counts = collections.defaultdict(int)
    for seq, pos in matches:
        motifs = set([x[0] for x in pos])
        for m in motifs:
            counts[m] += 1
    return counts

iupacdict = {
    'A': 'A',
    'C': 'C',
    'G': 'G',
    'T': 'T',
    'M': 'AC',
    'R': 'AG',
    'W': 'AT',
    'S': 'CG',
    'Y': 'CT',
    'K': 'GT',
    'V': 'ACG',
    'H': 'ACT',
    'D': 'AGT',
    'B': 'CGT',
    'X': 'ACGT',
    'N': 'ACGT'}

regexdict = dict(((x[1], x[0]) for x in iupacdict.items()))


def iupac2regex(pattern):
    '''convert iupac to regex pattern'''


def regex2iupac(pattern):
    '''convert regex to iupac pattern'''

    def _split(p):
        g = p.split("[")
        for c in g[0]:
            yield c
        for x in g[1:]:
            a, b = x.split("]")
            yield "".join(sorted(a))
            for c in b:
                yield c

    a = []
    for x in _split(pattern):
        a.append(regexdict[x])
    return "".join(a)
